# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


from gajim.common.const import FTState
from gajim.common.helpers import Observable


class FileTransfer(Observable):

    _state_descriptions: dict[FTState, str] = {}

    def __init__(self, account: str) -> None:
        Observable.__init__(self)

        self._account = account

        self._progress = 0

        self._state = FTState.INIT
        self._error_text: str = ''
        self._error_domain: str | None = None

    @property
    def account(self) -> str:
        return self._account

    @property
    def state(self) -> FTState:
        return self._state

    @property
    def filename(self) -> str:
        raise NotImplementedError

    @property
    def error_text(self) -> str:
        return self._error_text

    @property
    def error_domain(self) -> str | None:
        return self._error_domain

    def get_progress(self) -> int:
        return self._progress

    def set_progress(self, progress: int) -> None:
        self._progress = progress
        self.update_progress()

    def get_state_description(self) -> str:
        return self._state_descriptions.get(self._state, '')

    def set_preparing(self) -> None:
        self._state = FTState.PREPARING
        self.notify('state-changed', FTState.PREPARING)

    def set_encrypting(self) -> None:
        self._state = FTState.ENCRYPTING
        self.notify('state-changed', FTState.ENCRYPTING)

    def set_decrypting(self) -> None:
        self._state = FTState.DECRYPTING
        self.notify('state-changed', FTState.DECRYPTING)

    def set_started(self) -> None:
        self._state = FTState.STARTED
        self.notify('state-changed', FTState.STARTED)

    def set_error(self, domain: str, text: str = '') -> None:
        self._error_text = text
        self._error_domain = domain
        self._state = FTState.ERROR
        self.notify('state-changed', FTState.ERROR)
        self.disconnect_signals()

    def set_cancelled(self) -> None:
        self._state = FTState.CANCELLED
        self.notify('state-changed', FTState.CANCELLED)
        self.disconnect_signals()

    def set_in_progress(self) -> None:
        self._state = FTState.IN_PROGRESS
        self.notify('state-changed', FTState.IN_PROGRESS)

    def set_finished(self) -> None:
        self._state = FTState.FINISHED
        self.notify('state-changed', FTState.FINISHED)
        self.disconnect_signals()

    def update_progress(self) -> None:
        self.notify('progress')

    def cancel(self) -> None:
        self.notify('cancel')



# Copyright (C) 2005 Alex Mauer <hawke AT hawkesnest.net>
#                    Stéphan Kochen <stephan AT kochen.nl>
# Copyright (C) 2005-2006 Dimitur Kirov <dkirov AT gmail.com>
# Copyright (C) 2005-2007 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2005-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2006 Travis Shirk <travis AT pobox.com>
#                    Stefan Bethge <stefan AT lanpartei.de>
# Copyright (C) 2006-2007 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2007 Julien Pivotto <roidelapluie AT gmail.com>
# Copyright (C) 2007-2008 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2008 Jonathan Schleifer <js-gajim AT webkeks.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatOfflineParticipant
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.util.user_strings import get_uf_affiliation

from gajim.gtk.builder import get_builder
from gajim.gtk.util.misc import container_remove_all
from gajim.gtk.util.misc import iterate_children

log = logging.getLogger("gajim.gtk.tooltips")


class GCTooltip:
    def __init__(self) -> None:
        self._contact: GroupchatParticipant | GroupchatOfflineParticipant | None = None

    def clear_tooltip(self) -> None:
        self._contact = None

    def get_tooltip(
        self, contact: GroupchatParticipant | GroupchatOfflineParticipant
    ) -> tuple[bool, Gtk.Grid]:
        if not hasattr(self, "_ui"):
            self._ui = get_builder("groupchat_roster_tooltip.ui")

        if self._contact == contact:
            return True, self._ui.tooltip_grid

        self._populate_grid(contact)
        self._contact = contact
        return False, self._ui.tooltip_grid

    def _hide_grid_children(self) -> None:
        """
        Hide all Elements of the Tooltip Grid
        """
        for widget in iterate_children(self._ui.tooltip_grid):
            widget.set_visible(False)

    def _populate_grid(
        self, contact: GroupchatParticipant | GroupchatOfflineParticipant
    ) -> None:
        """
        Populate the Tooltip Grid with data of from the contact
        """
        self._hide_grid_children()

        self._ui.nick.set_text(contact.name)
        self._ui.nick.set_visible(True)

        # Status Message
        if contact.status:
            status = contact.status.strip()
            if status != "":
                self._ui.status.set_text(status)
                self._ui.status.set_visible(True)

        # JID
        if contact.real_jid is not None:
            self._ui.jid.set_text(str(contact.real_jid.bare))
            self._ui.jid.set_visible(True)

        # Affiliation
        if not contact.affiliation.is_none:
            uf_affiliation = get_uf_affiliation(contact.affiliation)
            uf_affiliation = _("%(owner_or_admin_or_member)s of this group chat") % {
                "owner_or_admin_or_member": uf_affiliation
            }
            self._ui.affiliation.set_text(uf_affiliation)
            self._ui.affiliation.set_visible(True)

        if contact.hats is not None:
            container_remove_all(self._ui.hats_box)

            for hat in contact.hats.get_hats()[:5]:
                # Limit to 5 hats
                hat_badge = Gtk.Box(spacing=6, halign=Gtk.Align.START)
                hat_badge.add_css_class("badge")
                hat_badge.add_css_class("badge-hat")
                if hat.hue is not None:
                    css_provider = Gtk.CssProvider()
                    css_provider.load_from_string(
                        f".badge-hat {{background-color: hsl({hat.hue}, 100%, 25%);}}"
                    )
                    context = hat_badge.get_style_context()
                    context.add_provider(
                        css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
                    )

                hat_badge_icon = Gtk.Image.new_from_icon_name("lucide-tag-symbolic")
                hat_badge.append(hat_badge_icon)

                hat_badge_label = Gtk.Label(
                    label=GLib.markup_escape_text(hat.title),
                    ellipsize=Pango.EllipsizeMode.END,
                    max_width_chars=20,
                    halign=Gtk.Align.START,
                )
                hat_badge.append(hat_badge_label)

                self._ui.hats_box.append(hat_badge)
                self._ui.hats_box.set_visible(True)

        # Avatar
        scale = self._ui.tooltip_grid.get_scale_factor()
        texture = contact.get_avatar(AvatarSize.TOOLTIP, scale)
        self._ui.avatar.set_pixel_size(AvatarSize.TOOLTIP)
        self._ui.avatar.set_from_paintable(texture)
        self._ui.avatar.set_visible(True)
        self._ui.fillelement.set_visible(True)

        app.plugin_manager.extension_point(
            "gc_tooltip_populate", self, contact, self._ui.tooltip_grid
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import Literal
from typing import TypedDict

import uuid

from nbxmpp.protocol import JID

from gajim.common.i18n import _


class _DEFAULT:
    pass


class _ACCOUNTDEFAULT:
    pass


HAS_APP_DEFAULT = _DEFAULT()
HAS_ACCOUNT_DEFAULT = _ACCOUNTDEFAULT()


BoolSettings = Literal[
    'always_english_wikipedia',
    'always_english_wiktionary',
    'ask_online_status',
    'autoaway',
    'autoxa',
    'chat_merge_consecutive_nickname',
    'check_for_update',
    'confirm_close_muc',
    'confirm_on_window_delete',
    'dev_force_bookmark_2',
    'enable_emoji_shortcodes',
    'enable_keepassxc_integration',
    'escape_key_closes',
    'enable_link_preview_default',
    'gc_enable_link_preview_default',
    'gc_notify_on_all_messages_private_default',
    'gc_notify_on_all_messages_public_default',
    'gc_notify_on_reaction_default',
    'gc_print_join_left_default',
    'gc_print_status_default',
    'hide_groupchat_occupants_list',
    'ignore_incoming_attention',
    'is_window_visible',
    'muc_prefer_direct_msg',
    'notification_preview_message',
    'notify_on_all_muc_messages',
    'notify_on_file_complete',
    'notify_on_reaction_default',
    'plugins_auto_update',
    'plugins_notify_after_update',
    'plugins_repository_enabled',
    'plugins_update_check',
    'positive_184_ack',
    'preview_allow_all_images',
    'preview_anonymous_muc',
    'preview_verify_https',
    'print_status_in_chats',
    'remote_control',
    'send_on_ctrl_enter',
    'show_help_start_chat',
    'show_in_taskbar',
    'show_notifications',
    'show_notifications_away',
    'show_send_message_button',
    'show_voice_message_button',
    'show_status_msgs_in_roster',
    'show_subject_on_join',
    'show_trayicon',
    'sort_by_show_in_muc',
    'sort_by_show_in_start_chat',
    'sounddnd',
    'sounds_on',
    'trayicon_notification_on_events',
    'use_keyring',
    'use_kib_mib',
    'use_speller',
    'use_stun_server',
    'use_urgency_hint',
    'video_see_self',
    'show_header_bar',
]

IntSettings = Literal[
    'audio_input_volume',
    'audio_output_volume',
    'autoawaytime',
    'autoxatime',
    'chat_handle_position',
    'dark_theme',
    'file_transfers_port',
    'gc_sync_threshold_private_default',
    'gc_sync_threshold_public_default',
    'groupchat_roster_width',
    'mainwin_height',
    'mainwin_width',
    'preview_max_file_size',
    'preview_size',
]

FloatSettings = Literal[
    'app_font_size',
]

StringSettings = Literal[
    'action_on_close',
    'additional_uri_schemes',
    'audio_input_device',
    'audio_output_device',
    'autoaway_message',
    'autoxa_message',
    'confirm_block',
    'date_format',
    'date_time_format',
    'dictionary_url',
    'gc_refer_to_nick_char',
    'global_proxy',
    'last_save_dir',
    'last_send_dir',
    'last_sounds_dir',
    'last_update_check',
    'latest_disco_addresses',
    'muc_highlight_words',
    'muclumbus_api_http_uri',
    'muclumbus_api_jid',
    'muclumbus_api_pref',
    'providers_list_url',
    'quick_reaction_emojis',
    'roster_theme',
    'search_engine',
    'show_main_window_on_startup',
    'speller_language',
    'stun_server',
    'time_format',
    'video_framerate',
    'video_input_device',
    'video_size',
]

AllSettings = Literal[BoolSettings, IntSettings, FloatSettings, StringSettings]
AllSettingsT = str | int | float | bool | list[str]

APP_SETTINGS: dict[str, str | int | float | bool | list[Any]] = {
    'additional_uri_schemes': '',
    'always_english_wikipedia': False,
    'always_english_wiktionary': True,
    'app_font_size': 1.0,
    'ask_online_status': False,
    'audio_input_device': 'autoaudiosrc ! volume name=gajim_vol',
    'audio_input_volume': 50,
    'audio_output_device': 'autoaudiosink',
    'audio_output_volume': 50,
    'autoaway': True,
    'autoaway_message': '',
    'autoawaytime': 5,
    'autoxa': True,
    'autoxa_message': '',
    'autoxatime': 15,
    'chat_handle_position': 350,
    'chat_merge_consecutive_nickname': True,
    'check_for_update': True,
    'confirm_block': '',
    'confirm_close_muc': True,
    'confirm_on_window_delete': True,
    'dark_theme': 2,
    'date_format': '%x',
    'date_time_format': '%c',
    'dev_force_bookmark_2': False,
    'dictionary_url': 'WIKTIONARY',
    'enable_emoji_shortcodes': True,
    'enable_keepassxc_integration': False,
    'escape_key_closes': False,
    'enable_link_preview_default': True,
    'file_transfers_port': 28011,
    'ft_add_hosts_to_send': '',
    'gc_enable_link_preview_default': True,
    'gc_notify_on_all_messages_private_default': True,
    'gc_notify_on_all_messages_public_default': False,
    'gc_notify_on_reaction_default': True,
    'gc_print_join_left_default': False,
    'gc_print_status_default': False,
    'gc_refer_to_nick_char': ',',
    'gc_sync_threshold_private_default': 0,
    'gc_sync_threshold_public_default': 1,
    'global_proxy': '',
    'groupchat_roster_width': 250,
    'hide_groupchat_occupants_list': False,
    'ignore_incoming_attention': False,
    'is_window_visible': True,
    'last_save_dir': '',
    'last_send_dir': '',
    'last_sounds_dir': '',
    'last_update_check': '',
    'latest_disco_addresses': '',
    'mainwin_height': 500,
    'mainwin_width': 1000,
    'action_on_close': 'hide',
    'muc_highlight_words': '',
    'muc_prefer_direct_msg': True,
    'muclumbus_api_http_uri': 'https://search.jabber.network/api/1.0/search',
    'muclumbus_api_jid': 'api@search.jabber.network',
    'muclumbus_api_pref': 'http',
    'notification_preview_message': True,
    'notify_on_all_muc_messages': False,
    'notify_on_file_complete': True,
    'notify_on_reaction_default': True,
    'plugins_auto_update': False,
    'plugins_notify_after_update': True,
    'plugins_repository_enabled': True,
    'plugins_update_check': True,
    'positive_184_ack': False,
    'preview_allow_all_images': False,
    'preview_anonymous_muc': False,
    'preview_max_file_size': 10485760,
    'preview_size': 400,
    'preview_verify_https': True,
    'providers_list_url': 'https://data.xmpp.net/providers/v2/providers-B.json',
    'print_status_in_chats': False,
    'quick_reaction_emojis': '👍,❤,🤣',
    'remote_control': False,
    'roster_theme': 'default',
    'search_engine': 'https://duckduckgo.com/?q=%s',
    'send_on_ctrl_enter': False,
    'show_header_bar': True,
    'show_help_start_chat': True,
    'show_in_taskbar': True,
    'show_main_window_on_startup': 'always',
    'show_notifications': True,
    'show_notifications_away': False,
    'show_send_message_button': False,
    'show_voice_message_button': True,
    'show_status_msgs_in_roster': True,
    'show_subject_on_join': True,
    'show_trayicon': True,
    'sort_by_show_in_muc': False,
    'sort_by_show_in_start_chat': True,
    'sounddnd': False,
    'sounds_on': True,
    'speller_language': '',
    'stun_server': '',
    'time_format': '%H:%M',
    'trayicon_notification_on_events': True,
    'use_keyring': True,
    'use_kib_mib': False,
    'use_speller': True,
    'use_stun_server': False,
    'use_urgency_hint': True,
    'video_framerate': '',
    'video_input_device': 'autovideosrc',
    'video_see_self': True,
    'video_size': '',
    'workspace_order': [],
}

BoolAccountSettings = Literal[
    'active',
    'anonymous_auth',
    'answer_receipts',
    'autoauth',
    'autoconnect',
    'autojoin_sync',
    'client_cert_encrypted',
    'confirm_unencrypted_connection',
    'enable_gssapi',
    'enable_security_labels',
    'ft_send_local_ips',
    'gc_send_marker_private_default',
    'gc_send_marker_public_default',
    'ignore_unknown_contacts',
    'publish_location',
    'publish_tune',
    'request_user_data',
    'restore_last_status',
    'savepass',
    'send_idle_time',
    'send_marker_default',
    'send_os_info',
    'sync_muc_blocks',
    'sync_with_global_status',
    'test_ft_proxies_on_startup',
    'update_timezone',
    'use_custom_host',
    'use_ft_proxies',
    'use_plain_connection',
    'omemo_blind_trust',
]


StringAccountSettings = Literal[
    'account_color',
    'account_label',
    'address',
    'avatar_sha',
    'client_cert',
    'custom_host',
    'custom_type',
    'default_workspace',
    'file_transfer_proxies',
    'filetransfer_preference',
    'gc_send_chatstate_default',
    'http_auth',
    'last_status',
    'last_status_msg',
    'password',
    'proxy',
    'resource',
    'roster_version',
    'send_chatstate_default',
    'subscription_request_msg',
]

IntAccountSettings = Literal[
    'chat_history_max_age',
    'custom_port',
]


AllAccountSettings = Literal[BoolAccountSettings,
                             IntAccountSettings,
                             StringAccountSettings]

AllAccountSettingsT = bool | int | str


BoolGroupChatSettings = Literal[
    'notify_on_all_messages',
    'print_join_left',
    'print_status',
    'send_marker',
    'enable_link_preview',
]

StringGroupChatSettings = Literal[
    'encryption',
    'mute_until',
    'speller_language',
    'send_chatstate',
]

IntGroupChatSettings = Literal[
    'sync_threshold',
]


AllGroupChatSettings = Literal[BoolGroupChatSettings,
                               IntGroupChatSettings,
                               StringGroupChatSettings]

AllGroupChatSettingsT = str | int | bool


BoolContactSettings = Literal[
    'send_marker',
    'enable_link_preview',
]

StringContactSettings = Literal[
    'encryption',
    'mute_until',
    'speller_language',
    'send_chatstate',
]

AllContactSettings = Literal[BoolContactSettings,
                             StringContactSettings]

AllContactSettingsT = str | bool


ACCOUNT_SETTINGS = {
    'account': {
        'account_color': 'rgb(85, 85, 85)',
        'account_label': '',
        'active': False,
        'address': '',
        'anonymous_auth': False,
        'answer_receipts': True,
        'autoauth': False,
        'autoconnect': True,
        'autojoin_sync': True,
        'avatar_sha': '',
        'chat_history_max_age': -1,
        'client_cert': '',
        'client_cert_encrypted': False,
        'confirm_unencrypted_connection': True,
        'custom_host': '',
        'custom_port': 5222,
        'custom_type': 'START TLS',
        'default_workspace': '',
        'enable_gssapi': False,
        'enable_security_labels': False,
        'encryption_default': '',
        'file_transfer_proxies': '',
        'filetransfer_preference': 'httpupload',
        'ft_send_local_ips': True,
        'gc_send_chatstate_default': 'composing_only',
        'gc_send_marker_private_default': True,
        'gc_send_marker_public_default': False,
        'http_auth': 'ask',
        'ignore_unknown_contacts': False,
        'last_status': 'online',
        'last_status_msg': '',
        'password': '',
        'proxy': '',
        'publish_location': False,
        'publish_tune': False,
        'request_user_data': True,
        'resource': 'gajim.$rand',
        'restore_last_status': False,
        'roster_version': '',
        'savepass': True,
        'send_chatstate_default': 'composing_only',
        'send_idle_time': True,
        'send_marker_default': True,
        'send_os_info': True,
        'subscription_request_msg': '',
        'sync_muc_blocks': True,
        'sync_with_global_status': True,
        'test_ft_proxies_on_startup': False,
        'update_timezone': True,
        'use_custom_host': False,
        'use_ft_proxies': False,
        'use_plain_connection': False,
        'omemo_blind_trust': True,
    },

    'contact': {
        'enable_link_preview': HAS_APP_DEFAULT,
        'encryption': HAS_ACCOUNT_DEFAULT,
        'mute_until': '',
        'send_chatstate': HAS_ACCOUNT_DEFAULT,
        'send_marker': HAS_ACCOUNT_DEFAULT,
        'speller_language': '',
    },

    'group_chat': {
        'enable_link_preview': HAS_APP_DEFAULT,
        'encryption': HAS_ACCOUNT_DEFAULT,
        'mute_until': '',
        'notify_on_all_messages': HAS_APP_DEFAULT,
        'print_join_left': HAS_APP_DEFAULT,
        'print_status': HAS_APP_DEFAULT,
        'send_chatstate': HAS_ACCOUNT_DEFAULT,
        'send_marker': HAS_ACCOUNT_DEFAULT,
        'speller_language': '',
        'sync_threshold': HAS_APP_DEFAULT,
    },
}

StringWorkspaceSettings = Literal[
    'avatar_sha',
    'color',
    'name',
]

AllWorkspaceSettings = Literal[StringWorkspaceSettings, 'chats']


class OpenChatSettingDetails(TypedDict):
    account: str
    jid: JID
    type: Literal["chat", "groupchat", "pm"]
    pinned: bool
    position: int


OpenChatsSettingT = list[OpenChatSettingDetails]
AllWorkspaceSettingsT = str | OpenChatsSettingT


class WorkspaceSettings(TypedDict):
    name: str
    color: str
    avatar_sha: str
    chats: OpenChatsSettingT


WORKSPACE_SETTINGS: WorkspaceSettings = {
    'name': _('My Workspace'),
    'color': '',
    'avatar_sha': '',
    'chats': [],
}


INITAL_WORKSPACE: dict[str, dict[str, WorkspaceSettings]] = {
    str(uuid.uuid4()): {}
}


PLUGIN_SETTINGS = {
    'active': False
}


PROXY_SETTINGS = {
    'type': 'socks5',
    'host': '',
    'port': 0,
    'useauth': False,
    'user': '',
    'pass': '',
}


PROXY_EXAMPLES = {
    'Tor': {
        'type': 'socks5',
        'host': 'localhost',
        'port': 9050
    },
    'Tor (Browser)': {
        'type': 'socks5',
        'host': 'localhost',
        'port': 9150
    },
}


DEFAULT_SOUNDEVENT_SETTINGS = {
    'attention_received': {
        'enabled': True,
        'path': 'attention.wav'
    },
    'first_message_received': {
        'enabled': True,
        'path': 'message1.wav'
    },
    'contact_connected': {
        'enabled': False,
        'path': 'connected.wav'
    },
    'contact_disconnected': {
        'enabled': False,
        'path': 'disconnected.wav'
    },
    'message_sent': {
        'enabled': False,
        'path': 'sent.wav'
    },
    'muc_message_highlight': {
        'enabled': True,
        'path': 'gc_message1.wav'
    },
    'muc_message_received': {
        'enabled': True,
        'path': 'message2.wav'
    },
    'incoming-call-sound': {
        'enabled': True,
        'path': 'call_incoming.wav'
    },
    'outgoing-call-sound': {
        'enabled': True,
        'path': 'call_outgoing.wav'
    }
}


ADVANCED_SETTINGS = {
    'app': {
        'additional_uri_schemes': _(
            'Clickable schemes in addition to the hard-coded list of '
            'IANA-registered ones. Space-separated, lower-case, no colons.'),
        'always_english_wikipedia': '',
        'always_english_wiktionary': '',
        'chat_merge_consecutive_nickname': _(
            'Show message meta data (avatar, nickname, timestamp) only once, '
            'if there are multiple messages from the same sender within a '
            'specific timespan.'),
        'confirm_block': _(
            'Show a confirmation dialog to block a contact? Empty string '
            'means never show the dialog.'),
        'confirm_close_muc': _('Ask before closing a group chat tab/window.'),
        'confirm_on_window_delete': _(
            'Ask before quitting when Gajim’s window is closed'),
        'date_format': 'https://docs.python.org/3/library/time.html#time.strftime',  # noqa: E501
        'date_time_format': 'https://docs.python.org/3/library/time.html#time.strftime',  # noqa: E501
        'dev_force_bookmark_2': _('Force Bookmark 2 usage'),
        'dictionary_url': _(
            'Either a custom URL with %%s in it (where %%s is the word/phrase)'
            ' or "WIKTIONARY" (which means use Wikitionary).'),
        'file_transfers_port': '',
        'ft_add_hosts_to_send': _(
            'List of send hosts (comma separated) in '
            'addition to local interfaces for file transfers (in case of '
            'address translation/port forwarding).'),
        'gc_notify_on_all_messages_private_default': '',
        'gc_notify_on_all_messages_public_default': '',
        'gc_refer_to_nick_char': _(
            'Character to add after nickname when using nickname completion '
            '(tab) in group chat.'),
        'groupchat_roster_width': _('Width of group chat roster in pixel'),
        'ignore_incoming_attention': _(
            'If enabled, Gajim will ignore incoming attention '
            'requests ("wizz").'),
        'muc_highlight_words': _(
            'A list of words (semicolon separated) that will be '
            'highlighted in group chats.'),
        'muclumbus_api_http_uri': '',
        'muclumbus_api_jid': '',
        'muclumbus_api_pref': _(
            'API Preferences. Possible values: "http", "iq"'),
        'notification_preview_message': _(
            'Preview new messages in notification popup?'),
        'notify_on_all_muc_messages': '',
        'plugins_repository_enabled': _(
            'If enabled, Gajim offers to download plugins hosted on gajim.org'),
        'providers_list_url': _(
            'Endpoint for retrieving a list of providers for sign up'),
        'quick_reaction_emojis': _(
            'Comma-separated list of emojis for quick reaction buttons'),
        'search_engine': '',
        'stun_server': _('STUN server to use when using Jingle'),
        'time_format': 'https://docs.python.org/3/library/time.html#time.strftime',  # noqa: E501
        'use_kib_mib': _(
            'IEC standard says KiB = 1024 bytes, KB = 1000 bytes.'),
        'use_stun_server': _(
            'If enabled, Gajim will try to use a STUN server when using Jingle.'
            ' The one in "stun_server" option, or the one given by '
            'the XMPP server.'),
        'use_urgency_hint': _(
            'If enabled, Gajim makes the window flash (the default behaviour '
            'in most Window Managers) when holding pending events.'),
    },
}



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import TYPE_CHECKING

import logging
from collections.abc import Iterator

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize

from gajim.gtk.chat_list_row import ChatListRow
from gajim.gtk.menus import get_workspace_menu
from gajim.gtk.sidebar_listbox import SideBarListBoxRow
from gajim.gtk.structs import ChatListEntryParam
from gajim.gtk.util.misc import get_listbox_row_count
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.misc import iterate_listbox_children
from gajim.gtk.util.window import open_window
from gajim.gtk.widgets import GajimPopover

if TYPE_CHECKING:
    # Simplifies testing
    from gajim.gtk.chat_list_stack import ChatListStack
    from gajim.gtk.chat_page import ChatPage

log = logging.getLogger("gajim.gtk.workspace_listbox")


@Gtk.Template(string=get_ui_string("workspace_listbox.ui"))
class WorkspaceListBox(Gtk.ListBox):
    __gtype_name__ = "WorkspaceListBox"

    def __init__(self) -> None:
        Gtk.ListBox.__init__(self)

        formats_workspace = Gdk.ContentFormats.new_for_gtype(SideBarListBoxRow)
        formats_chat_list_row = Gdk.ContentFormats.new_for_gtype(ChatListRow)
        formats = formats_workspace.union(formats_chat_list_row)

        drop_target = Gtk.DropTarget(formats=formats, actions=Gdk.DragAction.MOVE)
        drop_target.connect("drop", self._on_drop)
        self.add_controller(drop_target)

        self.set_sort_func(self._sort_func)

        self._popover_menu = GajimPopover(None)
        self._popover_menu.set_parent(self)

        self._workspaces: dict[str, SideBarListBoxRow] = {}

    def set_chat_page(self, chat_page: ChatPage) -> None:
        chat_list_stack = chat_page.get_chat_list_stack()
        chat_list_stack.connect("unread-count-changed", self._on_unread_count_changed)
        chat_list_stack.connect("chat-selected", self._on_chat_selected)

    def get_row_count(self) -> int:
        return get_listbox_row_count(self)

    def _iterate_rows(self) -> Iterator[SideBarListBoxRow]:
        for row in iterate_listbox_children(self):
            yield cast(SideBarListBoxRow, row)

    def _sort_func(self, row1: SideBarListBoxRow, row2: SideBarListBoxRow) -> int:
        return -1 if row1.index < row2.index else 1

    def _on_drop(
        self, _drop_target: Gtk.DropTarget, item: GObject.Value, _x: float, y: float
    ) -> bool:
        target_workspace = self.get_row_at_y(int(y))
        app.window.highlight_dnd_targets(self, False)

        if not item or not isinstance(target_workspace, SideBarListBoxRow):
            # Reject drop
            return False

        app.window.highlight_dnd_targets(item, False)

        if isinstance(item, SideBarListBoxRow):
            workspaces = list(self._iterate_rows())
            moved_workspace = workspaces.pop(item.index)
            workspaces.insert(target_workspace.get_index(), moved_workspace)

            for index, workspace in enumerate(workspaces):
                if workspace.item_id == "add-workspace":
                    continue
                workspace.index = index

            self.invalidate_sort()
            target_workspace.set_state_flags(Gtk.StateFlags.NORMAL, True)

            self.store_workspace_order()
            app.window.activate_workspace(item.item_id)
            return True

        if isinstance(item, ChatListRow):
            params = ChatListEntryParam(
                workspace_id=target_workspace.item_id,
                source_workspace_id=item.workspace_id,
                account=item.account,
                jid=item.jid,
            )
            app.window.activate_action(
                "win.move-chat-to-workspace", params.to_variant()
            )
            return True

        # Reject drop
        log.debug("Unknown item type dropped")
        return False

    def _on_unread_count_changed(
        self, _chat_list_stack: ChatListStack, workspace_id: str, count: int
    ) -> None:
        workspace = self._workspaces[workspace_id]
        workspace.set_unread_count(count)

    def _on_chat_selected(
        self, _chat_list_stack: ChatListStack, workspace_id: str, *args: Any
    ) -> None:
        self.activate_workspace(workspace_id)

    @Gtk.Template.Callback()
    def _on_row_activated(self, _listbox: Gtk.ListBox, row: SideBarListBoxRow) -> None:
        if row.item_id == "add-workspace":
            open_window("WorkspaceDialog")
        else:
            app.window.activate_workspace(row.item_id)

    def add_workspace(self, workspace_id: str) -> None:
        row = SideBarListBoxRow()
        row.set_workspace_id(workspace_id)
        row.set_secondary_callback(self._on_pressed)
        self._update_workspace_properties(row, workspace_id)

        row.index = self.get_row_count() - 1

        self._workspaces[workspace_id] = row

        self.insert(row, 0)

    def _update_workspace_properties(
        self, row: SideBarListBoxRow, workspace_id: str
    ) -> None:
        app.app.avatar_storage.invalidate_cache(workspace_id)
        scale = self.get_scale_factor()
        texture = app.app.avatar_storage.get_workspace_texture(
            workspace_id, AvatarSize.WORKSPACE, scale
        )
        row.set_from_paintable(texture)

        name = app.settings.get_workspace_setting(workspace_id, "name")
        row.set_tooltip_text(name)

    def store_workspace_order(self) -> None:
        workspaces = list(self._iterate_rows())
        order = [row.item_id for row in workspaces]
        order.remove("add-workspace")
        app.settings.set_app_setting("workspace_order", order)

    def remove_workspace(self, workspace_id: str) -> None:
        row = self._workspaces.pop(workspace_id)
        self.remove(row)

    def get_other_workspace(self, exclude_workspace_id: str) -> str | None:
        for workspace in self._workspaces.values():
            if workspace.item_id != exclude_workspace_id:
                return workspace.item_id
        return None

    def activate_workspace(self, workspace_id: str) -> None:
        row = cast(SideBarListBoxRow | None, self.get_selected_row())
        if row is not None and row.item_id == workspace_id:
            return

        row = self._workspaces[workspace_id]
        self.select_row(row)

    def activate_workspace_number(self, number: int) -> None:
        row = cast(SideBarListBoxRow | None, self.get_row_at_index(number))
        if row is not None and row.item_id != "add-workspace":
            app.window.activate_workspace(row.item_id)

    def get_active_workspace(self) -> str | None:
        row = cast(SideBarListBoxRow | None, self.get_selected_row())
        if row is None:
            return None
        return row.item_id

    def get_first_workspace(self) -> str:
        for row in self._iterate_rows():
            return row.item_id
        return ""

    def get_workspace_by_id(self, workspace_id: str) -> SideBarListBoxRow | None:
        for row in self._iterate_rows():
            if row.item_id == workspace_id:
                return row
        return None

    def update_avatar(self, workspace_id: str) -> None:
        row = self._workspaces[workspace_id]
        self._update_workspace_properties(row, workspace_id)

    def _on_pressed(
        self,
        gesture_click: Gtk.GestureClick,
        n_press: int,
        x: float,
        y: float,
        row: SideBarListBoxRow,
    ) -> None:
        res = row.translate_coordinates(self, x, y)
        if res is None:
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        x, y = res
        menu = get_workspace_menu(row.item_id)
        self._popover_menu.set_menu_model(menu)
        self._popover_menu.set_pointing_to_coord(x=x, y=y)
        self._popover_menu.popup()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal

import hashlib
import io
import logging
import math
import queue
import ssl
import threading
from collections.abc import Iterable
from dataclasses import dataclass
from functools import partial
from io import BytesIO
from pathlib import Path

import httpx
import truststore

from gajim.common.aes import AESGCMDecryptor
from gajim.common.aes import AESGCMEncryptor
from gajim.common.aes import AESKeyData
from gajim.common.enum import FTState

MIN_CHUNK_SIZE = 1024 * 100  # 100 KB
DEFAULT_MAX_CONTENT_LENGTH = 1024 * 1024 * 10  # 10 MB
NO_CONTENT_LENGTH_MAX_DOWNLOAD = 1024 * 1024  # 1 MB
USER_AGENT = "Gajim 2.x"

log = logging.getLogger("gajim.http")


@dataclass
class TransferState:
    id: str
    state: FTState
    total: int | None = None
    progress: int = 0


@dataclass
class TransferMetadata:
    id: str
    content_length: int | None
    content_type: str | None


@dataclass
class HTTPResult:
    hash_algo: str
    req_hash_value: str
    resp_hash_value: str
    content_length: int | None
    content_type: str | None
    content: bytes


class InvalidHash(Exception):
    pass


class MaxContentLengthExceeded(Exception):
    pass


class ContentTypeNotAllowed(Exception):
    pass


class HTTPStatusError(Exception):
    pass


class CancelledError(Exception):
    pass


class NonDecryptor:
    def decrypt(self, data: bytes) -> bytes:
        return data

    def finalize(self) -> bytes:
        return b""


class NonEncryptor:
    def encrypt(self, data: bytes) -> bytes:
        return data

    def finalize(self) -> bytes:
        return b""


def get_header_values(headers: httpx.Headers) -> tuple[int | None, str | None]:
    try:
        content_length = max(int(headers["Content-Length"]), 0)
    except Exception:
        content_length = None

    try:
        content_type = headers["Content-Type"]
    except Exception:
        content_type = None

    return content_length, content_type


def get_chunk_size(content_length: int | None, max_download_size: int = 0) -> int:
    if max_download_size > 0:
        return max_download_size

    if content_length is None:
        return MIN_CHUNK_SIZE
    return max(math.ceil(content_length / 100), MIN_CHUNK_SIZE)


def http_request(
    event: threading.Event,
    ft_id: str,
    method: Literal["GET", "POST", "PUT"],
    url: str,
    timeout: int,
    *,
    queue: queue.Queue[TransferState | TransferMetadata] | None = None,
    headers: dict[str, str] | None = None,
    params: dict[str, Any] | None = None,
    content_type: str | None = None,
    input_: Path | bytes = b"",
    with_req_progress: bool = False,
    output: Path | None = None,
    with_resp_progress: bool = False,
    max_content_length: int = DEFAULT_MAX_CONTENT_LENGTH,
    max_download_size: int = 0,
    allowed_content_types: Iterable[str] | None = None,
    hash_algo: str = "sha256",
    hash_value: str | None = None,
    encryption_data: AESKeyData | None = None,
    decryption_data: AESKeyData | None = None,
    proxy: str | None = None,
) -> HTTPResult:
    trust_env = True
    if proxy == "direct://":
        proxy = None
        trust_env = False

    ctx = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    client = httpx.Client(
        timeout=timeout,
        verify=ctx,
        http2=True,
        proxy=proxy,
        trust_env=trust_env,
        follow_redirects=True,
    )

    if queue is not None:
        queue.put(TransferState(id=ft_id, state=FTState.STARTED))

    if isinstance(input_, bytes):
        req_content_size = len(input_)
        input_file = io.BytesIO(input_)
    else:
        req_content_size = input_.stat().st_size
        input_file = input_.open(mode="rb")

    match encryption_data:
        case AESKeyData():
            encryptor = AESGCMEncryptor(encryption_data)
            req_content_size += 16
        case _:
            encryptor = NonEncryptor()

    default_headers = {
        "User-Agent": USER_AGENT,
    }

    if content_type is not None:
        default_headers.update(
            {
                "Content-Type": content_type,
                "Content-Length": str(req_content_size),
            }
        )

    if headers is not None:
        headers.update(default_headers)
    else:
        headers = default_headers

    req_hash_obj = hashlib.new(hash_algo)

    def _read_file_generator() -> Iterable[bytes]:
        if with_req_progress and queue is not None:
            queue.put(TransferState(id=ft_id, state=FTState.IN_PROGRESS))

        chunk_size = get_chunk_size(req_content_size)

        uploaded = 0
        while data := input_file.read(chunk_size):
            if event.is_set():
                input_file.close()
                raise CancelledError("HTTP Request was cancelled")

            req_hash_obj.update(data)
            data = encryptor.encrypt(data)
            uploaded += len(data)
            if with_req_progress and queue is not None:
                queue.put(
                    TransferState(
                        id=ft_id,
                        state=FTState.IN_PROGRESS,
                        total=req_content_size,
                        progress=uploaded,
                    )
                )
            yield data

        data = encryptor.finalize()
        uploaded += len(data)
        req_hash_obj.update(data)

        if with_req_progress and queue is not None:
            queue.put(
                TransferState(
                    id=ft_id,
                    state=FTState.IN_PROGRESS,
                    total=req_content_size,
                    progress=uploaded,
                )
            )

        input_file.close()
        yield data

    read_file_generator = None
    if input_:
        read_file_generator = _read_file_generator()

    req = client.build_request(
        method, url=url, content=read_file_generator, params=params, headers=headers
    )
    resp = client.send(req, stream=True)

    if event.is_set():
        raise CancelledError("HTTP Request was cancelled")

    try:
        resp.raise_for_status()
    except httpx.HTTPStatusError:
        # https://github.com/encode/httpx/issues/1990
        raise HTTPStatusError(f"{resp.status_code} {resp.reason_phrase}")

    content_length, content_type = get_header_values(resp.headers)

    if queue is not None:
        queue.put(
            TransferMetadata(
                id=ft_id, content_length=content_length, content_type=content_type
            )
        )

    if content_length == 0:
        return HTTPResult(
            hash_algo=hash_algo,
            req_hash_value=req_hash_obj.hexdigest(),
            resp_hash_value="",
            content_length=content_length,
            content_type=content_type,
            content=b"",
        )

    if content_length is not None:
        if max_content_length >= 0 and content_length > max_content_length:
            raise MaxContentLengthExceeded(content_length)

    if allowed_content_types is not None:
        if content_type is None or content_type not in allowed_content_types:
            raise ContentTypeNotAllowed(content_type)

    match decryption_data:
        case AESKeyData():
            decryptor = AESGCMDecryptor(decryption_data)
        case _:
            decryptor = NonDecryptor()

    if output is None:
        file_method = BytesIO
    else:
        file_method = partial(output.open, mode="wb")

    if with_resp_progress and queue is not None:
        queue.put(
            TransferState(
                id=ft_id, state=FTState.IN_PROGRESS, total=content_length, progress=0
            )
        )

    resp_hash_obj = hashlib.new(hash_algo)
    max_bytes_downloaded = content_length or NO_CONTENT_LENGTH_MAX_DOWNLOAD

    with file_method() as output_file:
        chunk_size = get_chunk_size(content_length, max_download_size)
        for data in resp.iter_bytes(chunk_size=chunk_size):
            if event.is_set():
                raise CancelledError("HTTP Request was cancelled")

            if (
                max_bytes_downloaded >= 0
                and resp.num_bytes_downloaded > max_bytes_downloaded
            ):
                raise OverflowError(
                    f"{resp.num_bytes_downloaded} > {max_bytes_downloaded}"
                )

            resp_hash_obj.update(data)
            output_file.write(decryptor.decrypt(data))
            if with_resp_progress and queue is not None:
                queue.put(
                    TransferState(
                        id=ft_id,
                        state=FTState.IN_PROGRESS,
                        total=content_length,
                        progress=resp.num_bytes_downloaded,
                    )
                )

            if resp.num_bytes_downloaded >= max_download_size > 0:
                break

        data = decryptor.finalize()
        output_file.write(data)
        resp_hash_obj.update(data)

        content = b""
        if isinstance(output_file, BytesIO):
            content = output_file.getvalue()

    resp_digest = resp_hash_obj.hexdigest()
    if hash_value is not None and resp_digest != hash_value:
        raise InvalidHash(f"{resp_digest} != {hash_value}")

    result = HTTPResult(
        hash_algo=hash_algo,
        req_hash_value=req_hash_obj.hexdigest(),
        resp_hash_value=resp_digest,
        content_length=content_length,
        content_type=content_type,
        content=content,
    )

    return result



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gtk

from gajim.common.i18n import _
from gajim.common.util.datetime import FIRST_LOCAL_DATETIME

from gajim.gtk.conversation.rows.base import BaseRow


class ScrollHintRow(BaseRow):
    def __init__(self, account: str) -> None:
        BaseRow.__init__(self, account, widget="label")
        self.set_selectable(False)
        self.set_activatable(False)

        self.type = "system"
        self.timestamp = FIRST_LOCAL_DATETIME

        self.add_css_class("conversation-system-row")

        self.label.set_halign(Gtk.Align.CENTER)
        self.label.set_hexpand(True)
        self.label.add_css_class("conversation-meta")
        self.grid.attach(self.label, 0, 1, 1, 1)

        self.set_history_complete(False)

    def set_history_complete(self, complete: bool) -> None:
        if complete:
            self.label.set_text(_("There is no more history"))
        else:
            self.label.set_text(_("Scroll up to load more chat history…"))

    def do_unroot(self) -> None:
        BaseRow.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from enum import Enum
from pathlib import Path

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.protocol import InvalidJid
from nbxmpp.protocol import JID
from nbxmpp.protocol import validate_resourcepart

from gajim.common import app
from gajim.common import types
from gajim.common.const import SimpleClientState
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.file_transfer_selector import FileTransferSelector
from gajim.gtk.groupchat_inviter import GroupChatInviter
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import container_remove_all


class FunctionMode(Enum):
    INVITE = "invite"
    SEND_FILE = "send-file"
    CHANGE_NICKNAME = "change-nickname"
    KICK = "kick"
    BAN = "ban"
    PASSWORD_REQUEST = "password-request"  # noqa: S105
    CAPTCHA_REQUEST = "captcha-request"
    CAPTCHA_ERROR = "captcha-error"
    JOIN_FAILED = "join-failed"
    CREATION_FAILED = "creation-failed"
    CONFIG_FAILED = "config-failed"


MUC_ERROR_CONDITION_TEXT = {
    "registration-required": _(
        "This group chat is members-only, but you are not on the members list. "
        "An authorized person needs to invite you."
    )
}


class ChatFunctionPage(Gtk.Box, SignalManager):
    __gsignals__ = {
        "finish": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
        "message": (GObject.SignalFlags.RUN_LAST, None, (str,)),
    }

    def __init__(self) -> None:
        Gtk.Box.__init__(
            self,
            orientation=Gtk.Orientation.VERTICAL,
            spacing=6,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        SignalManager.__init__(self)

        self.add_css_class("p-18")

        self._client: types.Client | None = None
        self._contact: types.ChatContactT | None = None
        self._mode: FunctionMode | None = None
        self._data: str | None = None
        self._ready_state = True

        self._widget: Gtk.Widget | None = None

        self._heading = Gtk.Label()
        self._heading.set_max_width_chars(30)
        self._heading.set_ellipsize(Pango.EllipsizeMode.END)
        self._heading.add_css_class("title-2")
        self.append(self._heading)

        self._subheading = Gtk.Label(
            max_width_chars=40,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD,
        )
        self._subheading.add_css_class("pb-18")
        self.append(self._subheading)

        self._content_box = Gtk.Box(halign=Gtk.Align.CENTER)
        self._content_box.add_css_class("pb-12")
        self.append(self._content_box)

        cancel_button = Gtk.Button(label=_("Cancel"))
        self._connect(cancel_button, "clicked", self._on_cancel_clicked)

        self._forget_button = Gtk.Button(label=_("Forget Group Chat"))
        self._forget_button.set_visible(False)
        self._forget_button.add_css_class("destructive-action")
        self._connect(self._forget_button, "clicked", self._on_forget_clicked)

        self._confirm_button = Gtk.Button()
        self._connect(self._confirm_button, "clicked", self._on_confirm_clicked)

        button_box = Gtk.Box(spacing=18, halign=Gtk.Align.CENTER)
        button_box.append(cancel_button)
        button_box.append(self._forget_button)
        button_box.append(self._confirm_button)

        self.append(button_box)

    def do_unroot(self) -> None:
        self._reset()
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def process_escape(self) -> None:
        close_control = self._mode in (
            FunctionMode.CREATION_FAILED,
            FunctionMode.CONFIG_FAILED,
            FunctionMode.CAPTCHA_ERROR,
        )
        self.emit("finish", close_control)

    def _reset(self) -> None:
        if self._client is not None:
            self._client.disconnect_all_from_obj(self)

        if self._widget is not None:
            self._disconnect_object(self._widget)
        container_remove_all(self._content_box)

        self._widget = None

        self._confirm_button.remove_css_class("destructive-action")
        self._confirm_button.remove_css_class("suggested-action")
        self._confirm_button.set_sensitive(False)
        self._forget_button.set_sensitive(True)
        self._forget_button.set_visible(False)

        self._ready_state = True

    def set_mode(
        self,
        contact: types.ChatContactT,
        mode: FunctionMode,
        data: str | None = None,
        files: list[str] | None = None,
    ) -> None:
        self._reset()

        self._contact = contact
        self._client = app.get_client(contact.account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._mode = mode
        self._data = data

        if isinstance(contact, BareContact) and contact.is_self:
            self._heading.set_text(_("Note to myself"))
        else:
            self._heading.set_text(self._contact.name)

        self._subheading.set_text(str(self._contact.jid))

        if mode == FunctionMode.INVITE:
            self._confirm_button.set_label(_("Invite"))
            self._confirm_button.add_css_class("suggested-action")
            self._widget = GroupChatInviter(str(contact.jid))
            self._widget.set_size_request(-1, 500)
            self._connect(self._widget, "listbox-changed", self._on_ready)
            self._widget.load_contacts()
            self._widget.grab_focus()

        elif mode == FunctionMode.SEND_FILE:
            self._confirm_button.set_label(_("Send Files"))
            self._confirm_button.add_css_class("suggested-action")
            self._widget = FileTransferSelector(self._contact, data)
            self._connect(self._widget, "changed", self._on_ready)
            if files is not None:
                self._widget.add_files(files)

        elif mode == FunctionMode.CHANGE_NICKNAME:
            self._confirm_button.set_label(_("Change"))
            self._confirm_button.add_css_class("suggested-action")
            self._widget = InputWidget(self._contact, mode)
            self._connect(self._widget, "changed", self._on_ready)

        elif mode == FunctionMode.KICK:
            self._confirm_button.set_label(_("Kick"))
            self._confirm_button.set_sensitive(True)
            self._confirm_button.add_css_class("destructive-action")
            self._widget = InputWidget(self._contact, mode, data)

        elif mode == FunctionMode.BAN:
            self._confirm_button.set_label(_("Ban"))
            self._confirm_button.set_sensitive(True)
            self._confirm_button.add_css_class("destructive-action")
            self._widget = InputWidget(self._contact, mode, data)

        elif mode == FunctionMode.PASSWORD_REQUEST:
            self._confirm_button.set_label(_("Join"))
            self._confirm_button.add_css_class("suggested-action")
            self._widget = InputWidget(self._contact, mode)
            self._connect(self._widget, "changed", self._on_ready)

        elif mode == FunctionMode.CAPTCHA_REQUEST:
            self._confirm_button.set_label(_("Join"))
            self._confirm_button.add_css_class("suggested-action")
            muc_data = self._client.get_module("MUC").get_muc_data(self._contact.jid)
            assert muc_data is not None
            form = muc_data.captcha_form
            assert form is not None
            options = {"no-scrolling": True, "entry-activates-default": True}
            self._widget = DataFormWidget(form, options=options)
            self._widget.set_valign(Gtk.Align.START)
            self._connect(self._widget, "is-valid", self._on_ready)

        elif mode == FunctionMode.CAPTCHA_ERROR:
            self._confirm_button.set_label(_("Try Again"))
            self._confirm_button.set_sensitive(True)
            self._confirm_button.add_css_class("suggested-action")
            self._widget = ErrorWidget(error_text=data)

        elif mode in (
            FunctionMode.JOIN_FAILED,
            FunctionMode.CREATION_FAILED,
            FunctionMode.CONFIG_FAILED,
        ):
            self._confirm_button.set_label(_("Try Again"))
            self._confirm_button.set_sensitive(True)
            self._confirm_button.add_css_class("suggested-action")
            if mode == FunctionMode.JOIN_FAILED:
                is_bookmark = self._client.get_module("Bookmarks").is_bookmark(
                    self._contact.jid
                )
                self._forget_button.set_visible(is_bookmark)

            self._widget = ErrorWidget(
                mode=mode,
                error_text=data,
            )

        assert self._widget is not None
        self._content_box.append(self._widget)
        if isinstance(self._widget, InputWidget):
            self._widget.focus()
        elif isinstance(self._widget, DataFormWidget):
            self._widget.focus_first_entry()
        elif mode == FunctionMode.SEND_FILE:
            self._confirm_button.grab_focus()

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, _state: SimpleClientState
    ) -> None:
        self._update_button_state()

    def _on_ready(self, _widget: Gtk.Widget, state: bool) -> None:
        self._ready_state = state
        self._update_button_state()

    def _update_button_state(self) -> None:
        assert self._contact is not None
        if app.account_is_connected(self._contact.account):
            self._confirm_button.set_sensitive(self._ready_state)
            self._forget_button.set_sensitive(True)
            return

        self._confirm_button.set_sensitive(False)
        self._forget_button.set_sensitive(False)

    def _on_confirm_clicked(self, _button: Gtk.Button) -> None:
        assert self._client is not None
        assert self._contact is not None

        if self._mode == FunctionMode.INVITE:
            assert isinstance(self._widget, GroupChatInviter)
            invitees = self._widget.get_invitees()
            for jid in invitees:
                self._invite(JID.from_string(jid))

        elif self._mode == FunctionMode.SEND_FILE:
            assert isinstance(self._widget, FileTransferSelector)
            if self._widget.transfer_resource_required():
                return

            self._send_files(self._widget.get_catalog())
            app.window.activate_action("win.input-focus")

        elif self._mode == FunctionMode.CHANGE_NICKNAME:
            assert isinstance(self._widget, InputWidget)
            nickname = self._widget.get_text()
            self._client.get_module("MUC").change_nick(self._contact.jid, nickname)

        elif self._mode == FunctionMode.KICK:
            assert isinstance(self._widget, InputWidget)
            reason = self._widget.get_text()
            self._client.get_module("MUC").set_role(
                self._contact.jid, self._data, "none", reason
            )

        elif self._mode == FunctionMode.BAN:
            assert isinstance(self._widget, InputWidget)
            reason = self._widget.get_text()
            self._client.get_module("MUC").set_affiliation(
                self._contact.jid,
                {self._data: {"affiliation": "outcast", "reason": reason}},
            )

        elif self._mode == FunctionMode.PASSWORD_REQUEST:
            assert isinstance(self._widget, InputWidget)
            password = self._widget.get_text()
            self._client.get_module("MUC").set_password(self._contact.jid, password)
            self._client.get_module("MUC").join(self._contact.jid)

        elif self._mode == FunctionMode.CAPTCHA_REQUEST:
            assert isinstance(self._widget, DataFormWidget)
            form_node = self._widget.get_submit_form()
            assert isinstance(form_node, SimpleDataForm)
            self._client.get_module("MUC").send_captcha(self._contact.jid, form_node)

        elif self._mode in (FunctionMode.JOIN_FAILED, FunctionMode.CAPTCHA_ERROR):
            self._client.get_module("MUC").join(self._contact.jid)

        self.emit("finish", False)

    def _on_cancel_clicked(self, _button: Gtk.Button) -> None:
        assert self._client is not None
        assert self._contact is not None

        connected = app.account_is_connected(self._contact.account)
        close_control = False

        if self._mode == FunctionMode.CAPTCHA_REQUEST:
            if connected:
                self._client.get_module("MUC").cancel_captcha(self._contact.jid)
            close_control = True

        elif self._mode == FunctionMode.PASSWORD_REQUEST:
            if connected:
                self._client.get_module("MUC").cancel_password_request(
                    self._contact.jid
                )
            close_control = True

        elif self._mode in (
            FunctionMode.CREATION_FAILED,
            FunctionMode.CONFIG_FAILED,
            FunctionMode.CAPTCHA_ERROR,
        ):
            close_control = True

        self.emit("finish", close_control)

    def _on_forget_clicked(self, _button: Gtk.Button) -> None:
        assert self._client is not None
        assert self._contact is not None
        self._client.get_module("Bookmarks").remove(self._contact.jid)
        self.emit("finish", True)

    def _invite(self, invited_jid: JID) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)
        client.get_module("MUC").invite(self._contact.jid, invited_jid)
        invited_contact = client.get_module("Contacts").get_contact(invited_jid)
        assert isinstance(invited_contact, BareContact)
        self.emit(
            "message",
            _("%s has been invited to this group chat") % invited_contact.name,
        )

    def _send_files(self, catalog: list[tuple[Path, str, JID]]) -> None:
        assert self._contact is not None
        client = app.get_client(self._contact.account)

        stack = app.window.get_chat_stack()
        if not stack.check_send_preconditons():
            return

        # catalog: list[(file Path, transfer method, recipient JID)]
        for path, method, _jid in catalog:
            if method == "httpupload":
                client.get_module("HTTPUpload").send_file(self._contact, path)
                continue

            # Send using Jingle
            # app.interface.instances["file_transfers"].send_file(
            #     self._contact.account, self._contact, jid, str(path)
            # )


class InputWidget(Gtk.Box, SignalManager):
    __gsignals__ = {
        "changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(
        self, contact: types.ChatContactT, mode: FunctionMode, data: str | None = None
    ) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=12)
        SignalManager.__init__(self)

        self._contact = contact
        self._mode = mode

        heading_label = Gtk.Label()
        heading_label.set_xalign(0)
        heading_label.add_css_class("title-3")
        self.append(heading_label)

        sub_label = Gtk.Label()
        sub_label.set_xalign(0)
        sub_label.add_css_class("dimmed")
        self.append(sub_label)

        self._entry = Gtk.Entry()
        self._entry.set_activates_default(True)
        self._entry.set_size_request(300, -1)
        self._connect(self._entry, "changed", self._on_entry_changed)
        self.append(self._entry)

        if mode == FunctionMode.CHANGE_NICKNAME:
            heading_label.set_text(_("Change Nickname"))
            sub_label.set_text(_("Enter your new nickname"))
            assert isinstance(self._contact, GroupchatContact)
            self._entry.set_text(self._contact.nickname or "")

        elif mode == FunctionMode.KICK:
            heading_label.set_text(_("Kick %s") % data)
            sub_label.set_text(_("Reason (optional)"))

        elif mode == FunctionMode.BAN:
            heading_label.set_text(_("Ban %s") % data)
            sub_label.set_text(_("Reason (optional)"))

        elif mode == FunctionMode.PASSWORD_REQUEST:
            heading_label.set_text(_("Password Required"))
            sub_label.set_text(_("Enter a password to join this chat"))
            self._entry.set_input_purpose(Gtk.InputPurpose.PASSWORD)
            self._entry.set_visibility(False)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)

    def _on_entry_changed(self, entry: Gtk.Entry) -> None:
        text = entry.get_text()

        if self._mode == FunctionMode.CHANGE_NICKNAME:
            assert isinstance(self._contact, GroupchatContact)
            if not text or text == self._contact.nickname:
                self.emit("changed", False)
                return
            try:
                validate_resourcepart(text)
            except InvalidJid:
                self.emit("changed", False)
                return

        self.emit("changed", bool(text))

    def focus(self) -> None:
        self._entry.grab_focus()

    def get_text(self) -> str:
        return self._entry.get_text()


class ErrorWidget(Gtk.Box):
    def __init__(
        self,
        mode: FunctionMode | None = None,
        error_text: str | None = None,
    ) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=12)

        image = Gtk.Image.new_from_icon_name("lucide-circle-x-symbolic")
        image.set_pixel_size(32)
        image.add_css_class("error")
        self.append(image)

        heading_text = _("An Error Occurred")
        if mode == FunctionMode.JOIN_FAILED:
            heading_text = _("Failed to Join Group Chat")
        elif mode == FunctionMode.CREATION_FAILED:
            heading_text = _("Failed to Create Group Chat")
        elif mode == FunctionMode.CONFIG_FAILED:
            heading_text = _("Failed to Configure Group Chat")

        heading = Gtk.Label(label=heading_text)
        heading.add_css_class("title-4")
        self.append(heading)

        if error_text is not None:
            error_text = MUC_ERROR_CONDITION_TEXT.get(error_text, error_text)
            error_label = Gtk.Label(
                max_width_chars=50,
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD,
                label=error_text,
            )
            self.append(error_label)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Generic
from typing import TypeVar

import locale
import logging

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp.client import Client as NBXMPPClient
from nbxmpp.errors import CancelledError
from nbxmpp.errors import is_error
from nbxmpp.errors import StanzaError
from nbxmpp.errors import TimeoutStanzaError
from nbxmpp.modules.muc.util import MucInfoResult
from nbxmpp.structs import DiscoInfo
from nbxmpp.structs import MuclumbusItem
from nbxmpp.structs import MuclumbusResult
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.client import Client
from gajim.common.const import AvatarSize
from gajim.common.const import Direction
from gajim.common.const import MUC_DISCO_ERRORS
from gajim.common.const import PresenceShowExt
from gajim.common.const import RFC5646_LANGUAGE_TAGS
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.util import as_task
from gajim.common.util.jid import validate_jid
from gajim.common.util.muc import get_group_chat_nick
from gajim.common.util.standards import get_rfc5646_lang
from gajim.common.util.status import compare_show
from gajim.common.util.text import to_one_line
from gajim.common.util.uri import parse_uri
from gajim.common.util.uri import XmppIri

from gajim.gtk.builder import get_builder
from gajim.gtk.chat_filter import ChatFilter
from gajim.gtk.chat_filter import ChatFilters
from gajim.gtk.chat_filter import ChatTypeFilter
from gajim.gtk.groupchat_info import GroupChatInfoScrolled
from gajim.gtk.groupchat_nick_chooser import GroupChatNickChooser
from gajim.gtk.menus import get_start_chat_menu
from gajim.gtk.menus import get_start_chat_row_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.icons import get_icon_theme
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import AccountBadge
from gajim.gtk.widgets import GajimPopover
from gajim.gtk.widgets import GroupBadgeBox
from gajim.gtk.widgets import IdleBadge
from gajim.gtk.window import GajimAppWindow

ContactT = BareContact | GroupchatContact
L = TypeVar("L", bound=type[GObject.Object])
V = TypeVar("V", bound=type[Gtk.Widget])


log = logging.getLogger("gajim.gtk.start_chat")


class StartChatDialog(GajimAppWindow):
    last_chat_filters = ChatFilters()

    def __init__(
        self, initial_jid: str | None = None, initial_message: str | None = None
    ) -> None:
        GajimAppWindow.__init__(
            self,
            name="StartChatDialog",
            title=_("Start / Join Chat"),
            default_height=600,
            default_width=550,
            header_bar=True,
        )

        self._parameter_form: MuclumbusResult | None = None
        self._keywords: list[str] = []
        self._destroyed = False
        self._search_stopped = False
        self._search_is_changed = False
        self._http_search_client: Client | None = None

        self._ui = get_builder("start_chat_dialog.ui")
        self.set_child(self._ui.stack)

        self._nick_chooser = GroupChatNickChooser()
        self._ui.join_box.prepend(self._nick_chooser)

        self._search_is_valid_jid = False

        self._new_contact_items: dict[str, ContactListItem] = {}
        self._accounts = app.get_enabled_accounts_with_labels()

        self._accounts_store = Gtk.ListStore(GdkPixbuf.Pixbuf, str, str)
        self._ui.account_view.set_model(self._accounts_store)

        self._contact_view = ContactListView()
        self._connect(self._contact_view, "activate", self._on_contact_item_activated)
        self._ui.contact_scrolled.set_child(self._contact_view)

        self._ui.settings_menu.set_menu_model(get_start_chat_menu())

        scale = self.get_scale_factor()
        self._add_accounts()
        self._add_contacts(scale)
        self._add_groupchats(scale)
        self._add_new_contact_items(scale)

        controller = Gtk.EventControllerKey(
            propagation_phase=Gtk.PropagationPhase.BUBBLE
        )
        text = self._ui.search_entry.get_delegate()
        assert isinstance(text, Gtk.Text)
        text.add_controller(controller)
        self._connect(controller, "key-pressed", self._on_search_key_pressed)

        self._connect(
            self._ui.infobar_close_button, "clicked", self._on_infobar_close_clicked
        )
        self._connect(self._ui.search_entry, "activate", self._on_search_activate)
        self._connect(self._ui.search_entry, "search-changed", self._on_search_changed)
        self._connect(
            self._ui.search_entry, "next-match", self._select_new_match, Direction.NEXT
        )
        self._connect(
            self._ui.search_entry,
            "previous-match",
            self._select_new_match,
            Direction.PREV,
        )
        self._connect(
            self._ui.search_entry,
            "stop-search",
            lambda *args: self._ui.search_entry.set_text(""),  # type: ignore
        )
        self._connect(
            self._ui.stack, "notify::visible-child-name", self._on_page_changed
        )
        self._connect(
            self._ui.global_search_toggle, "toggled", self._on_global_search_toggle
        )
        self._connect(self._ui.error_back_button, "clicked", self._on_back_clicked)
        self._connect(self._ui.join_button, "clicked", self._on_join_clicked)
        self._connect(self._ui.info_back_button, "clicked", self._on_back_clicked)
        self._connect(self._ui.account_view, "row-activated", self._on_select_clicked)
        self._connect(self._ui.account_back_button, "clicked", self._on_back_clicked)
        self._connect(
            self._ui.account_select_button, "clicked", self._on_select_clicked
        )

        self._global_search_view = GlobalSearch()
        self._connect(
            self._global_search_view, "activate", self._on_global_item_activated
        )
        self._connect(
            self._global_search_view,
            "global-search-progress",
            self._on_global_search_progress,
        )
        self._ui.global_scrolled.set_child(self._global_search_view)

        self._muc_info_box = GroupChatInfoScrolled()
        self._ui.info_box.prepend(self._muc_info_box)

        self._ui.infobar.set_reveal_child(app.settings.get("show_help_start_chat"))

        self._chat_filter = ChatFilter()
        self._connect(self._chat_filter, "filter-changed", self._on_chat_filter_changed)
        self._chat_filter.insert_after(self._ui.controls_box, self._ui.search_entry)

        self._connect(
            self.get_default_controller(), "key-pressed", self._on_key_pressed
        )

        self._initial_message: dict[str, str | None] = {}
        if initial_jid is not None:
            self._initial_message[initial_jid] = initial_message
            self._ui.search_entry.set_text(initial_jid)

        self._contact_view.set_loading_finished()

        action = Gio.SimpleAction.new_stateful(
            "sort-by-show",
            None,
            GLib.Variant.new_boolean(app.settings.get("sort_by_show_in_start_chat")),
        )
        self._connect(action, "change-state", self._on_sort_by_show_changed)
        self.add_action(action)

        self._ui.search_entry.grab_focus()
        log.debug("Loading dialog finished")

        self._chat_filter.set_filters(StartChatDialog.last_chat_filters)

    def _cleanup(self, *args: Any) -> None:
        del self._nick_chooser
        del self._global_search_view
        del self._muc_info_box
        del self._chat_filter
        del self._accounts_store
        del self._http_search_client
        self._new_contact_items.clear()
        self._destroyed = True
        app.cancel_tasks(self)

    def remove_row(self, account: str, jid: JID) -> None:
        # Used by forget-groupchat action
        self._contact_view.remove(account, jid)

    def _is_global_search_active(self) -> bool:
        return self._ui.list_stack.get_visible_child_name() == "global"

    def _get_active_view(self) -> ContactListView | GlobalSearch:
        if self._ui.list_stack.get_visible_child_name() == "global":
            return self._global_search_view
        return self._contact_view

    def _add_accounts(self) -> None:
        for account in self._accounts:
            self._accounts_store.append([None, *account])

    def _add_contacts(self, scale: int) -> None:
        log.debug("Loading contacts")
        show_account = len(self._accounts) > 1
        for account, _label in self._accounts:
            client = app.get_client(account)
            for jid, _data in client.get_module("Roster").iter():
                contact = client.get_module("Contacts").get_contact(jid)

                if isinstance(contact, GroupchatContact):
                    # Workaround if groupchats are in the roster
                    continue

                assert isinstance(contact, BareContact)

                item = ContactListItem(
                    account,
                    contact,
                    jid,
                    contact.name,
                    scale,
                    show_account,
                )

                self._contact_view.add(item)

            self_contact = client.get_module("Contacts").get_contact(
                client.get_own_jid().bare
            )
            assert isinstance(self_contact, BareContact)
            item = ContactListItem(
                account,
                self_contact,
                self_contact.jid,
                _("Note to myself"),
                scale,
                show_account,
            )
            self._contact_view.add(item)

        log.debug(
            "Loading contacts finished, model count %s", self._contact_view.get_count()
        )

    def _add_groupchats(self, scale: int) -> None:
        log.debug("Loading groupchats")
        show_account = len(self._accounts) > 1
        for account, _label in self._accounts:
            client = app.get_client(account)
            bookmarks = client.get_module("Bookmarks").bookmarks
            for bookmark in bookmarks:
                contact = client.get_module("Contacts").get_contact(
                    bookmark.jid, groupchat=True
                )
                assert isinstance(contact, GroupchatContact)

                item = ContactListItem(
                    account,
                    contact,
                    bookmark.jid,
                    contact.name,
                    scale,
                    show_account,
                    groupchat=True,
                )
                self._contact_view.add(item)

        log.debug(
            "Loading groupchats finished, model count %s",
            self._contact_view.get_count(),
        )

    def _add_new_contact_items(self, scale: int) -> None:
        for account, _label in self._accounts:
            show_account = len(self._accounts) > 1
            item = ContactListItem(account, None, None, None, scale, show_account)
            self._new_contact_items[account] = item
            self._contact_view.add(item)

    def _on_sort_by_show_changed(
        self, action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        action_state = action.get_state()
        assert action_state is not None
        new_state = not action_state.get_boolean()
        app.settings.set("sort_by_show_in_start_chat", new_state)
        action.set_state(GLib.Variant.new_boolean(new_state))
        self._contact_view.invalidate_sort()

    def _on_page_changed(self, stack: Gtk.Stack, _param: Any) -> None:
        if stack.get_visible_child_name() == "account":
            self._ui.account_view.grab_focus()

    def _on_contact_item_activated(
        self,
        _listbox: Gtk.ListView,
        position: int,
    ) -> None:
        item = self._contact_view.get_listitem(position)
        self._prepare_new_chat(item)

    def _on_global_item_activated(
        self,
        _listview: Gtk.ListView,
        position: int,
    ) -> None:
        self._select_muc()

    def _on_global_search_progress(
        self, _listview: Gtk.ListView, progressing: bool, results_count: int
    ) -> None:
        if progressing:
            self._ui.global_search_placeholder_stack.set_visible(True)
            self._ui.global_search_placeholder_stack.set_visible_child_name(
                "global-search-progress"
            )
            self._ui.global_search_results_label.set_text(
                _("Searching…\n%s results") % results_count
            )
            return

        if results_count == 0:
            self._ui.global_search_placeholder_stack.set_visible(True)
            self._ui.global_search_placeholder_stack.set_visible_child_name(
                "global-search-no-results"
            )
        else:
            self._ui.global_search_placeholder_stack.set_visible(False)

    def _select_muc(self) -> None:
        if len(self._accounts) > 1:
            self._ui.stack.set_visible_child_name("account")
        else:
            self._on_select_clicked()

    def _on_search_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Down:
            self._ui.search_entry.emit("next-match")
            return Gdk.EVENT_STOP

        if keyval == Gdk.KEY_Up:
            self._ui.search_entry.emit("previous-match")
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_search_activate(self, search_entry: Gtk.SearchEntry) -> None:
        if self._is_global_search_active() and self._search_is_changed:
            self._search_is_changed = False
            self._global_search_view.remove_all()
            self._start_search()
            return

        view = self._get_active_view()
        pos = view.get_selected()
        if pos == Gtk.INVALID_LIST_POSITION:
            return
        view.emit("activate", pos)

    def _on_search_changed(self, search_entry: Gtk.SearchEntry) -> None:
        self._search_is_changed = True
        self._show_search_entry_error(False)
        self._search_is_valid_jid = False

        if self._is_global_search_active():
            return

        search_text = search_entry.get_text()
        uri = parse_uri(search_text)
        if isinstance(uri, XmppIri):
            search_entry.set_text(str(uri.jid))
            return

        if search_text:
            try:
                validate_jid(search_text)
            except ValueError:
                self._show_search_entry_error(True)
            else:
                self._update_new_contact_items(search_text)
                self._search_is_valid_jid = True

        self._contact_view.set_search(search_text)

    def _on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Escape:
            if self._ui.stack.get_visible_child_name() == "progress":
                # Propagate to GajimAppWindow
                return Gdk.EVENT_PROPAGATE

            if self._ui.stack.get_visible_child_name() == "account":
                self._on_back_clicked()
                return Gdk.EVENT_STOP

            if self._ui.stack.get_visible_child_name() in ("error", "info"):
                self._ui.stack.set_visible_child_name("search")
                return Gdk.EVENT_STOP

            self._search_stopped = True
            self._ui.search_entry.grab_focus()
            self._global_search_view.remove_all()
            if self._ui.search_entry.get_text() != "":
                self._ui.search_entry.emit("stop-search")
                return Gdk.EVENT_STOP

            # Propagate to GajimAppWindow
            return Gdk.EVENT_PROPAGATE

        if keyval == Gdk.KEY_Return:
            if self._ui.stack.get_visible_child_name() == "progress":
                return Gdk.EVENT_STOP

            if self._ui.stack.get_visible_child_name() == "account":
                self._on_select_clicked()
                return Gdk.EVENT_STOP

            if self._ui.stack.get_visible_child_name() == "error":
                self._ui.stack.set_visible_child_name("search")
                return Gdk.EVENT_STOP

            if self._ui.stack.get_visible_child_name() == "info":
                self._on_join_clicked()
                return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_infobar_close_clicked(
        self,
        _button: Gtk.Button,
    ) -> None:
        self._ui.infobar.set_reveal_child(False)
        app.settings.set("show_help_start_chat", False)

    def _on_chat_filter_changed(self, chat_filter: ChatFilter) -> None:
        filters = chat_filter.get_filters()
        StartChatDialog.last_chat_filters = filters
        self._contact_view.set_chat_filter(chat_filter.get_filters())

    def _prepare_new_chat(self, item: ContactListItem) -> None:
        if item.jid is None:
            return

        if item.is_new:
            try:
                validate_jid(item.jid)
            except ValueError as error:
                self._show_error_page(str(error))
                return

            self._disco_info(item)
            return

        self._start_new_chat(item, groupchat=item.is_groupchat)

    def _start_new_chat(self, item: ContactListItem, *, groupchat: bool) -> None:
        jid = JID.from_string(item.jid)
        if groupchat:
            if not app.account_is_available(item.account):
                self._show_error_page(
                    _("You can not join a group chat unless you are connected.")
                )
                return

            if app.window.chat_exists(item.account, jid):
                app.window.select_chat(item.account, jid)
                self.close()
                return

            self._disco_muc(item.account, jid, request_vcard=item.is_new)

        else:
            initial_message = self._initial_message.get(item.jid)
            app.window.add_chat(
                item.account, jid, "chat", select=True, message=initial_message
            )
            self.close()

    def _disco_info(self, item: ContactListItem) -> None:
        if not app.account_is_available(item.account):
            self._show_error_page(_("You are offline"))
            return

        self._ui.stack.set_visible_child_name("progress")
        client = app.get_client(item.account)
        client.get_module("Discovery").disco_info(
            item.jid, callback=self._disco_info_received, user_data=item, timeout=10
        )

    def _disco_info_received(self, task: Task) -> None:
        item = cast(ContactListItem, task.get_user_data())
        try:
            result = cast(DiscoInfo, task.finish())
        except StanzaError as error:
            contact_conditions = [
                "service-unavailable",  # Prosody
                "subscription-required",  # ejabberd
                "feature-not-implemented",  # transports/bridges
            ]
            if error.condition in contact_conditions:
                # These error conditions are the result of
                # querying contacts without subscription
                self._start_new_chat(item, groupchat=False)
                return

            # Handle other possible errors
            self._show_error_page(to_user_string(error))
            return
        except TimeoutStanzaError:
            # We reached the 10s timeout and we cannot
            # assume which kind contact this is.
            self._show_error_page(_("This address is not reachable."))
            return

        groupchat = False
        if result.is_muc and not result.jid.is_domain:
            # This is mostly a fix for the MUC protocol, there is no
            # way to differentiate between a MUC service and room.
            # Except the MUC XEP defines rooms should have a localpart.
            groupchat = True

        self._start_new_chat(item, groupchat=groupchat)

    def _disco_muc(self, account: str, jid: JID, request_vcard: bool) -> None:
        self._ui.stack.set_visible_child_name("progress")
        client = app.get_client(account)
        client.get_module("Discovery").disco_muc(
            jid,
            request_vcard=request_vcard,
            allow_redirect=True,
            timeout=10,
            callback=self._muc_disco_info_received,
            user_data=account,
        )

    def _muc_disco_info_received(self, task: Task) -> None:
        try:
            result = cast(MucInfoResult, task.finish())
        except (StanzaError, TimeoutStanzaError) as error:
            self._set_error(error)
            return

        account = task.get_user_data()

        if result.info.is_muc:
            self._muc_info_box.set_account(account)
            self._muc_info_box.set_from_disco_info(result.info)
            self._nick_chooser.set_text(get_group_chat_nick(account, result.info.jid))
            self._ui.stack.set_visible_child_name("info")

        else:
            self._set_error_from_code("not-muc-service")

    def _set_error(self, error: StanzaError | TimeoutStanzaError) -> None:
        if isinstance(error, TimeoutStanzaError):
            text = _("This address is not reachable.")
        else:
            text = MUC_DISCO_ERRORS.get(error.condition, to_user_string(error))
            if error.condition == "gone":
                reason = error.get_text(get_rfc5646_lang())
                if reason:
                    text = f"{text}:\n{reason}"
        self._show_error_page(text)

    def _set_error_from_code(self, error_code: str) -> None:
        self._show_error_page(MUC_DISCO_ERRORS[error_code])

    def _show_error_page(self, text: str) -> None:
        self._ui.error_label.set_text(str(text))
        self._ui.stack.set_visible_child_name("error")

    def _on_join_clicked(self, *args: Any) -> None:
        account = self._muc_info_box.get_account()
        jid = self._muc_info_box.get_jid()
        nickname = self._nick_chooser.get_text()
        assert account
        app.window.show_add_join_groupchat(account, str(jid), nickname=nickname)

        self.close()

    def _on_back_clicked(self, *args: Any) -> None:
        self._ui.stack.set_visible_child_name("search")

    def _on_select_clicked(self, *args: Any) -> None:
        model, iter_ = self._ui.account_view.get_selection().get_selected()
        if iter_ is not None:
            account = model[iter_][1]
        elif len(self._accounts) == 1:
            account = self._accounts[0][0]
        else:
            return

        item = self._global_search_view.get_selected_item()
        if item is None:
            return

        if not app.account_is_available(account):
            self._show_error_page(
                _("You can not join a group chat unless you are connected.")
            )
            return

        jid = JID.from_string(item.jid)
        self._disco_muc(account, jid, request_vcard=True)

    def _on_global_search_toggle(self, button: Gtk.ToggleButton) -> None:
        self._ui.search_entry.grab_focus()
        image = cast(Gtk.Image, button.get_child())
        if button.get_active():
            self._chat_filter.reset()
            self._chat_filter.set_sensitive(False)
            image.add_css_class("accent")
            self._ui.list_stack.set_visible_child_name("global")

            self._ui.global_search_placeholder_stack.set_visible_child_name(
                "global-search-hint"
            )
            self._ui.global_search_placeholder_stack.set_visible(True)

            if self._ui.search_entry.get_text():
                self._start_search()
                self._global_search_view.grab_focus()
        else:
            self._chat_filter.set_sensitive(True)
            self._ui.search_entry.set_text("")
            image.remove_css_class("accent")
            self._ui.list_stack.set_visible_child_name("contacts")
            self._global_search_view.remove_all()

    def _show_search_entry_error(self, state: bool):
        self._ui.search_error_box.set_visible(state)

    def _update_new_contact_items(self, search_text: str) -> None:
        for item in self._new_contact_items.values():
            item.props.jid = JID.from_string(search_text)

    def _select_new_match(self, _entry: Gtk.Entry, direction: Direction) -> None:
        if self._is_global_search_active():
            self._global_search_view.select(direction)
        else:
            self._contact_view.select(direction)

    def _start_search(self) -> None:
        self._search_stopped = False
        accounts = app.get_connected_accounts()
        if not accounts:
            return
        client = app.get_client(accounts[0]).connection

        text = self._ui.search_entry.get_text().strip()
        if not text:
            return

        self._global_search_view.start_search()

        if app.settings.get("muclumbus_api_pref") == "http":
            self._start_http_search(app.get_client(accounts[0]), text)
        else:
            self._start_iq_search(client, text)

    @as_task
    def _start_iq_search(self, client: NBXMPPClient, text: str):
        _task = yield  # noqa: F841

        if self._parameter_form is None:
            result = yield client.get_module("Muclumbus").request_parameters(
                app.settings.get("muclumbus_api_jid")
            )

            self._process_search_result(result, parameters=True)

            self._parameter_form = result
            self._parameter_form.type_ = "submit"

        self._parameter_form.vars["q"].value = text

        result = yield client.get_module("Muclumbus").set_search(
            app.settings.get("muclumbus_api_jid"), self._parameter_form
        )

        self._process_search_result(result)

        while not result.end:
            result = yield client.get_module("Muclumbus").set_search(
                app.settings.get("muclumbus_api_jid"),
                self._parameter_form,
                items_per_page=result.max,
                after=result.last,
            )

            self._process_search_result(result)

        self._global_search_view.end_search()

    def _start_http_search(self, client: Client, text: str) -> None:
        self._keywords = text.split(" ")
        self._http_search_client = client
        client.get_module("HttpMucSearch").search(
            self._keywords, None, self._on_http_result
        )

    def _on_http_result(self, result: MuclumbusResult) -> None:
        try:
            self._process_search_result(result)
        except CancelledError:
            return

        if result.end:
            self._global_search_view.end_search()
            return

        assert self._http_search_client is not None
        self._http_search_client.get_module("HttpMucSearch").search(
            self._keywords, result.last, self._on_http_result
        )

    def _process_search_result(
        self, result: MuclumbusResult, parameters: bool = False
    ) -> None:
        if self._search_stopped:
            raise CancelledError

        if is_error(result):
            assert isinstance(result, StanzaError)
            self._ui.global_search_placeholder_stack.set_visible(False)
            self._show_error_page(to_user_string(result))
            raise result

        if parameters:
            return

        for item in result.items:
            self._global_search_view.add(item)


class BaseListView(Generic[L, V], Gtk.ListView, SignalManager):
    _selection_model: Gtk.SingleSelection
    _filter_model: Gtk.FilterListModel

    def __init__(self, list_type: L, view_type: V) -> None:
        Gtk.ListView.__init__(self)
        SignalManager.__init__(self)

        self._model = Gio.ListStore(item_type=list_type)

        factory = Gtk.SignalListItemFactory()
        self._connect(factory, "setup", self._on_factory_setup, view_type)
        self._connect(factory, "bind", self._on_factory_bind)
        self._connect(factory, "unbind", self._on_factory_unbind)
        self.set_factory(factory)

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem, view_type: V
    ) -> None:
        list_item.set_child(view_type())

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = list_item.get_child()
        obj = list_item.get_item()
        view_item.bind(obj)

    @staticmethod
    def _on_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        view_item = list_item.get_child()
        view_item.unbind()

    def add(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def remove_all(self) -> None:
        self._model.remove_all()

    def get_listitem(self, position: int) -> L:
        return self._filter_model.get_item(position)

    def get_selected_item(self) -> L | None:
        return self._selection_model.get_selected_item()

    def get_selected(self) -> int:
        return self._selection_model.get_selected()

    def select(self, direction: Direction) -> None:
        selected_pos = self._selection_model.get_selected()
        if selected_pos == Gtk.INVALID_LIST_POSITION:
            return

        if direction == Direction.NEXT:
            selected_pos += 1
        else:
            selected_pos -= 1

        if not 0 <= selected_pos < self._selection_model.get_n_items():
            return

        self.scroll_to(selected_pos, Gtk.ListScrollFlags.SELECT)


class ContactListView(BaseListView[type["ContactListItem"], type["ContactViewItem"]]):
    def __init__(self) -> None:
        BaseListView.__init__(self, ContactListItem, ContactViewItem)

        self._chat_filters = ChatFilters()
        self._scroll_id = None
        self._search_string_list: list[str] = []

        self._sorter = Gtk.CustomSorter.new(sort_func=self._sort_func)
        self._sort_model = Gtk.SortListModel(sorter=self._sorter)

        self._custom_filter = Gtk.CustomFilter.new(self._filter_func)

        self._filter_model = Gtk.FilterListModel(
            model=self._sort_model, filter=self._custom_filter
        )
        self._connect(
            self._filter_model, "items-changed", self._on_filter_items_changed
        )

        self._selection_model = Gtk.SingleSelection(model=self._filter_model)

        self.set_model(self._selection_model)

    def do_unroot(self) -> None:
        # The filter func needs to be unset before calling do_unroot (see #12213)
        self._custom_filter.set_filter_func(None)
        self._sorter.set_sort_func(None)
        Gtk.ListView.do_unroot(self)
        self._disconnect_all()
        app.check_finalize(self._model)
        app.check_finalize(self._filter_model)
        app.check_finalize(self._sort_model)
        app.check_finalize(self._selection_model)
        app.check_finalize(self._custom_filter)
        del self._model
        del self._filter_model
        del self._sort_model
        del self._selection_model
        del self._custom_filter
        del self._sorter
        app.check_finalize(self)

    def set_loading_finished(self) -> None:
        self._sort_model.set_model(self._model)

    def get_count(self) -> int:
        return self._model.get_n_items()

    def invalidate_sort(self) -> None:
        self._sorter.changed(Gtk.SorterChange.DIFFERENT)

    def _on_filter_items_changed(
        self, filter_model: Gtk.FilterListModel, _pos: int, _removed: int, _added: int
    ) -> None:
        # Cancel any active source at first so we dont have
        # multiple timeouts running

        if self._scroll_id is not None:
            GLib.source_remove(self._scroll_id)
            self._scroll_id = None

        # If the first item is already selected or
        # no items are in the model we dont need to trigger a scroll

        if self._selection_model.get_selected() == 0 or filter_model.get_n_items() == 0:
            return

        def _scroll_to() -> None:
            self._scroll_id = None
            self.scroll_to(0, Gtk.ListScrollFlags.SELECT)

        self._scroll_id = GLib.timeout_add(50, _scroll_to)

    def _filter_func(self, item: ContactListItem) -> bool:
        account = self._chat_filters.account
        if account is not None and account != item.account:
            return False

        if item.is_new:
            return True

        for search_string in self._search_string_list:
            if search_string not in item.search_string:
                return False

        group = self._chat_filters.group
        if group is not None and group not in item.groups:
            return False

        type_ = self._chat_filters.type
        if type_ == ChatTypeFilter.ALL:
            return True

        is_groupchat = item.is_groupchat
        if type_ == ChatTypeFilter.CHAT and not is_groupchat:
            return True

        return type_ == ChatTypeFilter.GROUPCHAT and is_groupchat

    @staticmethod
    def _sort_func(
        obj1: ContactListItem,
        obj2: ContactListItem,
        _user_data: object | None,
    ) -> int:
        if obj1.is_new != obj2.is_new:
            return 1 if obj1.is_new else -1

        if obj1.is_groupchat != obj2.is_groupchat:
            return 1 if obj1.is_groupchat else -1

        if obj1.is_self != obj2.is_self:
            return 1 if obj1.is_self else -1

        if app.settings.get("sort_by_show_in_start_chat"):
            res = compare_show(obj1.show, obj2.show)
            if res != 0:
                return res

        return locale.strcoll(obj1.name.lower(), obj2.name.lower())

    def add(self, item: ContactViewItem) -> None:
        self._model.append(item)

    def remove(self, account: str, jid: JID) -> None:
        for item in cast(list[ContactListItem], self._model):
            if item.account != account or item.jid != jid:
                continue
            success, pos = self._model.find(item)
            if success:
                self._model.remove(pos)
            break

    def set_search(self, text: str) -> None:
        self._search_string_list = text.lower().split()
        self._custom_filter.changed(Gtk.FilterChange.DIFFERENT)

    def set_chat_filter(self, value: ChatFilters) -> None:
        if self._chat_filters == value:
            return
        self._chat_filters = value
        self._custom_filter.changed(Gtk.FilterChange.DIFFERENT)


class ContactListItem(GObject.Object):
    __gtype_name__ = "ContactListItem"

    account = GObject.Property(type=str)
    account_visible = GObject.Property(type=bool, default=False)
    jid = GObject.Property(type=str)
    name = GObject.Property(type=str)
    groups = GObject.Property(type=object)
    is_groupchat = GObject.Property(type=bool, default=False)
    is_new = GObject.Property(type=bool, default=False)
    is_self = GObject.Property(type=bool, default=False)
    avatar_paintable = GObject.Property(type=Gdk.Paintable)
    idle = GObject.Property(type=object)
    show = GObject.Property(type=object)
    status = GObject.Property(type=str)
    status_visible = GObject.Property(type=bool, default=False)
    search_string = GObject.Property(type=str)
    menu = GObject.Property(type=Gio.Menu)

    def __init__(
        self,
        account: str,
        contact: ContactT | None,
        jid: JID | None,
        name: str | None,
        scale: int,
        account_visible: bool,
        groupchat: bool = False,
    ) -> None:
        name = name or _("Start / Join Chat")

        idle = None
        status = ""
        groups = []
        show = PresenceShowExt.OFFLINE
        is_self = False
        if contact is not None and not groupchat:
            assert isinstance(contact, BareContact)
            groups = sorted(contact.groups)
            status = to_one_line(contact.status)
            idle = contact.idle_datetime
            show = contact.show
            is_self = contact.is_self

        menu = get_start_chat_row_menu(account, jid)

        is_new = jid is None

        avatar_paintable = None
        if is_new:
            theme = get_icon_theme()
            avatar_paintable = theme.lookup_icon(
                "lucide-user-plus-symbolic",
                None,
                AvatarSize.START_CHAT,
                scale,
                Gtk.TextDirection.NONE,
                0,  # type: ignore
            )

        else:
            assert isinstance(contact, BareContact | GroupchatContact)
            avatar_paintable = contact.get_avatar(AvatarSize.START_CHAT, scale)

        search_string = "|".join((name, str(jid))).lower()

        super().__init__(
            account=account,
            account_visible=account_visible,
            jid=jid,
            name=name,
            idle=idle,
            is_groupchat=groupchat,
            is_new=jid is None,
            is_self=is_self,
            show=show,
            status=status,
            status_visible=bool(status),
            groups=groups,
            menu=menu,
            avatar_paintable=avatar_paintable,
            search_string=search_string,
        )

    def __repr__(self) -> str:
        return f"ContactListItem: {self.props.account} - {self.props.jid}"


@Gtk.Template(string=get_ui_string("contact_view_item.ui"))
class ContactViewItem(Gtk.Grid, SignalManager):
    __gtype_name__ = "ContactViewItem"

    _avatar: Gtk.Image = Gtk.Template.Child()
    _name_label: Gtk.Label = Gtk.Template.Child()
    _address_label: Gtk.Label = Gtk.Template.Child()
    _status_label: Gtk.Label = Gtk.Template.Child()
    _idle_badge: IdleBadge = Gtk.Template.Child()
    _account_badge: AccountBadge = Gtk.Template.Child()
    _group_badge_box: GroupBadgeBox = Gtk.Template.Child()
    _menu: GajimPopover = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Grid.__init__(self)
        SignalManager.__init__(self)

        self.__bindings: list[GObject.Binding] = []

        gesture_secondary_click = Gtk.GestureClick(
            button=Gdk.BUTTON_SECONDARY, propagation_phase=Gtk.PropagationPhase.BUBBLE
        )
        self._connect(gesture_secondary_click, "pressed", self._popup_menu)
        self.add_controller(gesture_secondary_click)

    def bind(self, obj: ContactListItem) -> None:
        bind_spec = [
            ("name", self._name_label, "label"),
            ("jid", self._address_label, "label"),
            ("status", self._status_label, "label"),
            ("status_visible", self._status_label, "visible"),
            ("avatar_paintable", self._avatar, "paintable"),
            ("idle", self._idle_badge, "idle"),
            ("account", self._account_badge, "account"),
            ("account_visible", self._account_badge, "visible"),
            ("groups", self._group_badge_box, "groups"),
            ("menu", self._menu, "menu-model"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

        if obj.is_new:
            self._avatar.add_css_class("dimmed")

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Grid.do_unroot(self)
        app.check_finalize(self)

    def _popup_menu(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        if self._menu.get_menu_model() is None:
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        self._menu.set_pointing_to_coord(x, y)
        self._menu.popup()


class GlobalSearch(
    BaseListView[type["GlobalListItem"], type["GlobalViewItem"]],
    Gtk.ListView,
    SignalManager,
):
    __gsignals__ = {
        "global-search-progress": (GObject.SignalFlags.RUN_FIRST, None, (bool, int)),
    }

    def __init__(self) -> None:
        BaseListView.__init__(self, GlobalListItem, GlobalViewItem)
        SignalManager.__init__(self)

        self._results_count = 0

        self._selection_model = Gtk.SingleSelection(model=self._model)
        self.set_model(self._selection_model)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.ListView.do_unroot(self)
        del self._model
        app.check_finalize(self)

    def start_search(self) -> None:
        self._results_count = 0
        self.emit("global-search-progress", True, 0)

    def end_search(self) -> None:
        self.emit("global-search-progress", False, self._results_count)

    def add(self, item: MuclumbusItem) -> None:
        self._results_count += 1
        self.emit("global-search-progress", True, self._results_count)
        self._model.append(GlobalListItem(item=item))


class GlobalListItem(GObject.Object):
    __gtype_name__ = "GlobalListItem"

    jid = GObject.Property(type=str)
    name = GObject.Property(type=str)
    nusers = GObject.Property(type=str)
    description = GObject.Property(type=str)
    language_visible = GObject.Property(type=bool, default=False)
    language = GObject.Property(type=str)
    language_code = GObject.Property(type=str)

    def __init__(self, item: MuclumbusItem) -> None:
        jid = JID.from_string(item.jid)
        name = item.name or jid.localpart or str(jid)
        language = RFC5646_LANGUAGE_TAGS.get(item.language, item.language) or None

        super().__init__(
            jid=item.jid,
            name=name,
            nusers=item.nusers,
            description=item.description,
            language_visible=bool(language is not None),
            language=_("Language: %s") % language,
            language_code=item.language.upper()[:2],
        )

    def __repr__(self) -> str:
        return f"GlobalListItem: {self.props.jid} {self.props.name}"


@Gtk.Template(string=get_ui_string("global_view_item.ui"))
class GlobalViewItem(Gtk.Box):
    __gtype_name__ = "GlobalViewItem"

    _name_box: Gtk.Box = Gtk.Template.Child()
    _name_label: Gtk.Label = Gtk.Template.Child()
    _description_label: Gtk.Label = Gtk.Template.Child()
    _language_box: Gtk.Box = Gtk.Template.Child()
    _language_label: Gtk.Label = Gtk.Template.Child()
    _users_count_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Box.__init__(self)
        self.__bindings: list[GObject.Binding] = []

    def bind(self, obj: GlobalListItem) -> None:
        bind_spec: list[tuple[str, Gtk.Widget, str]] = [
            ("name", self._name_label, "label"),
            ("description", self._description_label, "label"),
            ("jid", self._name_box, "tooltip_text"),
            ("language_visible", self._language_box, "visible"),
            ("language", self._language_box, "tooltip_text"),
            ("language_code", self._language_label, "label"),
            ("nusers", self._users_count_label, "label"),
        ]

        for source_prop, child_widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, child_widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import ParamSpec
from typing import TypeVar

import logging
from collections.abc import Callable
from functools import wraps
from time import monotonic

from gi.repository import GLib

P = ParamSpec("P")
R = TypeVar("R")


log = logging.getLogger("gajim.c.util.decorators")


def cache_with_ttl(ttl: int = 2) -> Any:
    class Result:
        __slots__ = ("timeout", "value")

        def __init__(self, value: Any, timeout: float) -> None:
            self.value = value
            self.timeout = timeout

    def decorator(func: Any) -> Any:
        def cached_func(*args: Any) -> Result:
            value = func(*args)
            timeout = monotonic() + ttl
            return Result(value, timeout)

        @wraps(func)
        def wrapper(module: Any, *args: Any) -> Any:
            ttl_cache = module.get_ttl_cache()
            result = ttl_cache.get((module, *args))
            if result is not None:
                if result.timeout < monotonic():
                    result.value = func(module, *args)
                    result.timeout = monotonic() + ttl

                return result.value

            result = cached_func(module, *args)
            ttl_cache[(module, *args)] = result
            return result.value

        return wrapper

    return decorator


def catch_exceptions(func: Callable[P, R]) -> Callable[P, R | None]:
    @wraps(func)
    def func_wrapper(*args: P.args, **kwargs: P.kwargs) -> R | None:
        try:
            result = func(*args, **kwargs)
        except Exception as error:
            log.exception(error)
            return None

        return result

    return func_wrapper


def event_filter(filter_: Any):
    def event_filter_decorator(func: Any) -> Any:
        @wraps(func)
        def func_wrapper(self: Any, event: Any, *args: Any, **kwargs: Any) -> Any:
            for attr in filter_:
                if "=" in attr:
                    attr1, attr2 = attr.split("=")
                else:
                    attr1, attr2 = attr, attr
                try:
                    if getattr(event, attr1) != getattr(self, attr2):
                        return None
                except AttributeError:
                    if getattr(event, attr1) != getattr(self, f"_{attr2}"):
                        return None

            return func(self, event, *args, **kwargs)

        return func_wrapper

    return event_filter_decorator


def delay_execution(milliseconds: int) -> Any:
    # Delay the first call for `milliseconds`
    # ignore all other calls while the delay is active
    def delay_execution_decorator(func: Any) -> Any:
        @wraps(func)
        def func_wrapper(*args: Any, **kwargs: Any) -> Any:
            def timeout_wrapper():
                func(*args, **kwargs)
                delattr(func_wrapper, "source_id")

            if hasattr(func_wrapper, "source_id"):
                return
            func_wrapper.source_id = GLib.timeout_add(  # type: ignore
                milliseconds, timeout_wrapper
            )

        return func_wrapper

    return delay_execution_decorator



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import GObject
from gi.repository import Gtk

from gajim.common.i18n import _


class MessageSelection(Gtk.Grid):
    __gsignals__ = {
        "copy": (GObject.SignalFlags.RUN_LAST, None, ()),
        "cancel": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self) -> None:
        Gtk.Grid.__init__(
            self,
            row_spacing=18,
            column_spacing=6,
            visible=False,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.END,
        )

        self.add_css_class("toolbar")
        self.add_css_class("osd")

        label = Gtk.Label(
            label=_("Click messages to select them\n(Ctrl + Double Click to deselect)")
        )
        self.attach(label, 0, 0, 2, 1)

        cancel_button = Gtk.Button(label=_("Cancel"))
        cancel_button.connect("clicked", self._on_cancel_clicked)
        self.attach(cancel_button, 0, 1, 1, 1)

        copy_button = Gtk.Button(label=_("Copy Text"))
        copy_button.add_css_class("suggested-action")
        copy_button.connect("clicked", self._on_copy_clicked)
        self.attach(copy_button, 1, 1, 1, 1)

    def _on_copy_clicked(self, _button: Gtk.Button) -> None:
        self.emit("copy")

    def _on_cancel_clicked(self, _button: Gtk.Button) -> None:
        self.emit("cancel")



# This is a port of um-crop-area.c from GNOME’s 'Cheese' application, see
# https://gitlab.gnome.org/GNOME/cheese/-/blob/3.34.0/libcheese/um-crop-area.c
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast

import logging
from enum import IntEnum
from enum import unique
from pathlib import Path

import cairo
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import Display
from gajim.common.i18n import _
from gajim.common.util.image import scale_with_ratio
from gajim.common.util.uri import get_file_path_from_uri

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.filechoosers import AvatarFileChooserButton
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import gi_gui_package_version

log = logging.getLogger("gajim.gtk.avatar_selector")


@unique
class Loc(IntEnum):
    OUTSIDE = 0
    INSIDE = 1
    TOP = 2
    TOP_LEFT = 3
    TOP_RIGHT = 4
    BOTTOM = 5
    BOTTOM_LEFT = 6
    BOTTOM_RIGHT = 7
    LEFT = 8
    RIGHT = 9


@unique
class Range(IntEnum):
    BELOW = 0
    LOWER = 1
    BETWEEN = 2
    UPPER = 3
    ABOVE = 4


class AvatarSelector(Gtk.Box, SignalManager):
    def __init__(self) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL, spacing=12)
        SignalManager.__init__(self)

        self.add_css_class("avatar-selector")

        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        self._connect(drop_target, "accept", self._on_drop_accept)
        self._connect(drop_target, "drop", self._on_file_drop)
        self.add_controller(drop_target)

        self._crop_area = CropArea()
        self._crop_area.set_vexpand(True)
        self.append(self._crop_area)

        self._file_chooser_button = AvatarFileChooserButton(label=_("Load Image"))
        self._file_chooser_button.set_halign(Gtk.Align.CENTER)
        self._connect(self._file_chooser_button, "path-picked", self._on_path_picked)
        self.append(self._file_chooser_button)

        self._helper_label = Gtk.Label(label=_("…or drop it here"))
        self._helper_label.add_css_class("bold")
        self._helper_label.add_css_class("dimmed")
        self._helper_label.set_vexpand(True)
        self.append(self._helper_label)

    def do_unroot(self, *args: Any) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        app.check_finalize(self)
        del self._file_chooser_button
        del self._crop_area

    def reset(self) -> None:
        self._crop_area.reset()
        self._file_chooser_button.reset()
        self._file_chooser_button.set_visible(True)
        self._helper_label.set_visible(True)

    def prepare_crop_area(self, path: str) -> None:
        pixbuf = self._get_pixbuf_from_path(path)
        if pixbuf is None:
            log.info("Could not load from path %s", path)
            return

        self._crop_area.set_pixbuf(pixbuf)
        self._file_chooser_button.set_visible(False)
        self._helper_label.set_visible(False)
        self._crop_area.set_visible(True)

    def _on_path_picked(
        self, _button: AvatarFileChooserButton, paths: list[Path]
    ) -> None:
        if not paths:
            return

        self.prepare_crop_area(str(paths[0]))

    def _on_drop_accept(self, _target: Gtk.DropTarget, drop: Gdk.Drop) -> bool:
        # DND on X11 freezes due to a GTK bug:
        # https://dev.gajim.org/gajim/gajim/-/issues/12313
        if app.is_display(Display.X11) and not gi_gui_package_version("Gtk>=4.20.1"):
            return False

        formats = drop.get_formats()
        return bool(formats.contain_gtype(Gdk.FileList))

    def _on_file_drop(
        self, _target: Gtk.DropTarget, value: Gdk.FileList | None, _x: float, _y: float
    ) -> bool:
        if value is None:
            log.debug("Drop received, but value is None")
            return False

        files = value.get_files()
        if not files:
            return False

        path = get_file_path_from_uri(files[0].get_uri())
        if path is None:
            return False

        self.prepare_crop_area(str(path))
        return True

    @staticmethod
    def _get_pixbuf_from_path(path: str) -> GdkPixbuf.Pixbuf | None:
        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(path)
            return pixbuf
        except GLib.Error as err:
            log.error("Unable to load file %s: %s", path, str(err))
            return None

    def get_prepared(self) -> bool:
        return bool(self._crop_area.get_pixbuf())

    @staticmethod
    def _scale_for_publish(
        pixbuf: GdkPixbuf.Pixbuf,
    ) -> tuple[GdkPixbuf.Pixbuf, int, int]:
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        if width > AvatarSize.PUBLISH or height > AvatarSize.PUBLISH:
            # Scale only down, never up
            width, height = scale_with_ratio(AvatarSize.PUBLISH, width, height)
            scaled_pixbuf = pixbuf.scale_simple(
                width, height, GdkPixbuf.InterpType.BILINEAR
            )
            assert scaled_pixbuf is not None
            return scaled_pixbuf, width, height
        return pixbuf, width, height

    def get_avatar_texture(self) -> tuple[Gdk.Texture, int, int] | None:
        pixbuf = self._crop_area.get_pixbuf()
        if pixbuf is None:
            return None
        scaled, width, height = self._scale_for_publish(pixbuf)

        return Gdk.Texture.new_for_pixbuf(scaled), width, height

    def get_avatar_bytes(self) -> tuple[bool, bytes | None, int, int]:
        pixbuf = self._crop_area.get_pixbuf()
        if pixbuf is None:
            return False, None, 0, 0
        scaled, width, height = self._scale_for_publish(pixbuf)

        success, data = scaled.save_to_bufferv("png", [], [])
        return success, data, width, height


class CropArea(Gtk.DrawingArea, SignalManager):
    def __init__(self) -> None:
        Gtk.DrawingArea.__init__(self, visible=False)
        SignalManager.__init__(self)

        self._image = Gdk.Rectangle()
        self._crop = Gdk.Rectangle()
        self._pixbuf: GdkPixbuf.Pixbuf | None = None
        self._browse_pixbuf: GdkPixbuf.Pixbuf | None = None
        self._color_shifted_pixbuf: GdkPixbuf.Pixbuf | None = None
        self._current_cursor: Any | None = None

        self._scale = 1.0
        self._image.x = 0
        self._image.y = 0
        self._image.width = 0
        self._image.height = 0
        self._active_region = Loc.OUTSIDE
        self._last_press_x = -1
        self._last_press_y = -1
        self._base_width = 10
        self._base_height = 10
        self._aspect = 1.0

        self.set_size_request(self._base_width, self._base_height)

        self.set_draw_func(self._on_draw)

        gesture_primary_click = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        self._connect(gesture_primary_click, "pressed", self._on_button_press)
        self._connect(gesture_primary_click, "released", self._on_button_release)
        self.add_controller(gesture_primary_click)

        motion_controller = Gtk.EventControllerMotion()
        self._connect(motion_controller, "motion", self._on_motion)
        self.add_controller(motion_controller)

    def do_unroot(self, *args: Any) -> None:
        self._disconnect_all()
        self.set_draw_func(None)
        Gtk.DrawingArea.do_unroot(self)
        app.check_finalize(self)

    def set_min_size(self, width: int, height: int) -> None:
        self._base_width = width
        self._base_height = height
        self.set_size_request(self._base_width, self._base_height)

        if self._aspect > 0:
            self._aspect = self._base_width / self._base_height

    def set_contstrain_aspect(self, constrain: bool) -> None:
        if constrain:
            self._aspect = self._base_width / self._base_height
        else:
            self._aspect = -1

    def reset(self) -> None:
        self._browse_pixbuf = None
        self.set_visible(False)

    def set_pixbuf(self, pixbuf: GdkPixbuf.Pixbuf) -> None:
        self._browse_pixbuf = pixbuf
        width = pixbuf.get_width()
        height = pixbuf.get_height()

        self._crop.width = 2 * self._base_width
        self._crop.height = 2 * self._base_height
        self._crop.x = int(abs((width - self._crop.width) / 2))
        self._crop.y = int(abs((height - self._crop.height) / 2))

        self._scale = 1.0
        self._image.x = 0
        self._image.y = 0
        self._image.width = 0
        self._image.height = 0

        self._update_pixbufs()

        self.queue_draw()

    def get_pixbuf(self) -> GdkPixbuf.Pixbuf | None:
        if self._browse_pixbuf is None:
            return None

        width = self._browse_pixbuf.get_width()
        height = self._browse_pixbuf.get_height()
        width = min(self._crop.width, width - self._crop.x)
        height = min(self._crop.height, height - self._crop.y)

        if width <= 0 or height <= 0:
            return None

        return GdkPixbuf.Pixbuf.new_subpixbuf(
            self._browse_pixbuf, self._crop.x, self._crop.y, width, height
        )

    def _on_draw(
        self,
        _drawing_area: Gtk.DrawingArea,
        context: cairo.Context[cairo.ImageSurface],
        _width: int,
        _height: int,
    ) -> None:
        if self._browse_pixbuf is None:
            return

        self._update_pixbufs()

        if self._pixbuf is None:
            avatar_selector = cast(AvatarSelector, self.get_parent())
            avatar_selector.reset()
            InformationAlertDialog(
                _("Error Loading Image"), _("Selected image could not be loaded.")
            )
            return

        width = self._pixbuf.get_width()
        height = self._pixbuf.get_height()
        crop = self._crop_to_widget()

        ix = self._image.x
        iy = self._image.y

        assert self._color_shifted_pixbuf is not None
        Gdk.cairo_set_source_pixbuf(context, self._color_shifted_pixbuf, ix, iy)
        context.rectangle(ix, iy, width, crop.y - iy)
        context.rectangle(ix, crop.y, crop.x - ix, crop.height)
        context.rectangle(
            crop.x + crop.width, crop.y, width - crop.width - (crop.x - ix), crop.height
        )
        context.rectangle(
            ix, crop.y + crop.height, width, height - crop.height - (crop.y - iy)
        )
        context.fill()

        Gdk.cairo_set_source_pixbuf(context, self._pixbuf, ix, iy)
        context.rectangle(crop.x, crop.y, crop.width, crop.height)
        context.fill()

        if self._active_region != Loc.OUTSIDE:
            context.set_source_rgb(150, 150, 150)
            context.set_line_width(1.0)
            x1 = crop.x + crop.width / 3.0
            x2 = crop.x + 2 * crop.width / 3.0
            y1 = crop.y + crop.height / 3.0
            y2 = crop.y + 2 * crop.height / 3.0

            context.move_to(x1 + 0.5, crop.y)
            context.line_to(x1 + 0.5, crop.y + crop.height)

            context.move_to(x2 + 0.5, crop.y)
            context.line_to(x2 + 0.5, crop.y + crop.height)

            context.move_to(crop.x, y1 + 0.5)
            context.line_to(crop.x + crop.width, y1 + 0.5)

            context.move_to(crop.x, y2 + 0.5)
            context.line_to(crop.x + crop.width, y2 + 0.5)
            context.stroke()

        context.set_source_rgb(1, 1, 1)
        context.set_line_width(1.0)

        context.rectangle(
            crop.x + 0.5, crop.y + 0.5, crop.width - 1.0, crop.height - 1.0
        )
        context.stroke()

        context.set_source_rgb(1, 1, 1)
        context.set_line_width(2.0)
        context.rectangle(
            crop.x + 2.0, crop.y + 2.0, crop.width - 4.0, crop.height - 4.0
        )
        context.stroke()

    def _on_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        if self._browse_pixbuf is None:
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        crop = self._crop_to_widget()

        self._last_press_x = (x - self._image.x) / self._scale
        self._last_press_y = (y - self._image.y) / self._scale
        self._active_region = self._find_location(crop, x, y)
        self.queue_draw()

    def _on_button_release(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        if self._browse_pixbuf is None:
            return

        self._crop_to_widget()
        self._last_press_x = -1
        self._last_press_y = -1
        self._active_region = Loc.OUTSIDE

        self.queue_draw()

    def _on_motion(
        self,
        _event_controller: Gtk.EventControllerMotion,
        x: float,
        y: float,
    ) -> None:
        # pylint: disable=too-many-boolean-expressions
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        if self._browse_pixbuf is None:
            return

        self._update_cursor(x, y)

        self._crop_to_widget()
        self.queue_draw()

        pb_width = self._browse_pixbuf.get_width()
        pb_height = self._browse_pixbuf.get_height()

        x_coord = int((x - self._image.x) / self._scale)
        y_coord = int((y - self._image.y) / self._scale)

        delta_x = int(x_coord - self._last_press_x)
        delta_y = int(y_coord - self._last_press_y)
        self._last_press_x = x_coord
        self._last_press_y = y_coord

        left = int(self._crop.x)
        right = int(self._crop.x + self._crop.width - 1)
        top = int(self._crop.y)
        bottom = int(self._crop.y + self._crop.height - 1)

        center_x = float((left + right) / 2.0)
        center_y = float((top + bottom) / 2.0)

        if self._active_region == Loc.INSIDE:
            width = right - left + 1
            height = bottom - top + 1

            left += delta_x
            right += delta_x
            top += delta_y
            bottom += delta_y

            left = max(left, 0)
            top = max(top, 0)
            right = min(right, pb_width)
            bottom = min(bottom, pb_height)

            adj_width = int(right - left + 1)
            adj_height = int(bottom - top + 1)
            if adj_width != width:
                if delta_x < 0:
                    right = left + width - 1
                else:
                    left = right - width + 1

            if adj_height != height:
                if delta_y < 0:
                    bottom = top + height - 1
                else:
                    top = bottom - height + 1

        elif self._active_region == Loc.TOP_LEFT:
            if self._aspect < 0:
                top = y_coord
                left = x_coord
            elif y_coord < self._eval_radial_line(
                center_x, center_y, left, top, x_coord
            ):
                top = y_coord
                new_width = float((bottom - top) * self._aspect)
                left = right - new_width
            else:
                left = x_coord
                new_height = float((right - left) / self._aspect)
                top = bottom - new_height

        elif self._active_region == Loc.TOP:
            top = y_coord
            if self._aspect > 0:
                new_width = float((bottom - top) * self._aspect)
                right = left + new_width

        elif self._active_region == Loc.TOP_RIGHT:
            if self._aspect < 0:
                top = y_coord
                right = x_coord
            elif y_coord < self._eval_radial_line(
                center_x, center_y, right, top, x_coord
            ):
                top = y_coord
                new_width = float((bottom - top) * self._aspect)
                right = left + new_width
            else:
                right = x_coord
                new_height = float((right - left) / self._aspect)
                top = bottom - new_height

        elif self._active_region == Loc.LEFT:
            left = x_coord
            if self._aspect > 0:
                new_height = float((right - left) / self._aspect)
                bottom = top + new_height

        elif self._active_region == Loc.BOTTOM_LEFT:
            if self._aspect < 0:
                bottom = y_coord
                left = x_coord
            elif y_coord < self._eval_radial_line(
                center_x, center_y, left, bottom, x_coord
            ):
                left = x_coord
                new_height = float((right - left) / self._aspect)
                bottom = top + new_height
            else:
                bottom = y_coord
                new_width = float((bottom - top) * self._aspect)
                left = right - new_width

        elif self._active_region == Loc.RIGHT:
            right = x_coord
            if self._aspect > 0:
                new_height = float((right - left) / self._aspect)
                bottom = top + new_height

        elif self._active_region == Loc.BOTTOM_RIGHT:
            if self._aspect < 0:
                bottom = y_coord
                right = x_coord
            elif y_coord < self._eval_radial_line(
                center_x, center_y, right, bottom, x_coord
            ):
                right = x_coord
                new_height = float((right - left) / self._aspect)
                bottom = top + new_height
            else:
                bottom = y_coord
                new_width = float((bottom - top) * self._aspect)
                right = left + new_width

        elif self._active_region == Loc.BOTTOM:
            bottom = y_coord
            if self._aspect > 0:
                new_width = float((bottom - top) * self._aspect)
                right = left + new_width
        else:
            return

        min_width = int(self._base_width / self._scale)
        min_height = int(self._base_height / self._scale)

        width = right - left + 1
        height = bottom - top + 1
        if self._aspect < 0:
            left = max(left, 0)
            top = max(top, 0)
            right = min(right, pb_width)
            bottom = min(bottom, pb_height)

            width = right - left + 1
            height = bottom - top + 1

            if self._active_region in (Loc.LEFT, Loc.TOP_LEFT, Loc.BOTTOM_LEFT):
                if width < min_width:
                    left = right - min_width
            elif self._active_region in (Loc.RIGHT, Loc.TOP_RIGHT, Loc.BOTTOM_RIGHT):
                if width < min_width:
                    right = left + min_width

            if self._active_region in (Loc.TOP, Loc.TOP_LEFT, Loc.TOP_RIGHT):
                if height < min_height:
                    top = bottom - min_height
            elif self._active_region in (Loc.BOTTOM, Loc.BOTTOM_LEFT, Loc.BOTTOM_RIGHT):
                if height < min_height:
                    bottom = top + min_height

        else:
            if (
                left < 0
                or top < 0
                or right > pb_width
                or bottom > pb_height
                or width < min_width
                or height < min_height
            ):
                left = self._crop.x
                right = self._crop.x + self._crop.width - 1
                top = self._crop.y
                bottom = self._crop.y + self._crop.height - 1

        self._crop.x = int(left)
        self._crop.y = int(top)
        self._crop.width = int(right - left + 1)
        self._crop.height = int(bottom - top + 1)
        self._crop_to_widget()

        self.queue_draw()

    def _update_pixbufs(self) -> None:
        allocated_width = self.get_width()
        allocated_height = self.get_height()

        assert self._browse_pixbuf
        width = self._browse_pixbuf.get_width()
        height = self._browse_pixbuf.get_height()

        scale = allocated_height / float(height)
        if scale * width > allocated_width:
            scale = allocated_width / float(width)

        dest_width = int(width * scale)
        dest_height = int(height * scale)

        if dest_width == 0 or dest_height == 0:
            log.warning("Image width or height is zero")
            return

        if (
            self._pixbuf is None
            or self._pixbuf.get_width() != allocated_width
            or self._pixbuf.get_height() != allocated_height
        ):
            self._pixbuf = GdkPixbuf.Pixbuf.new(
                GdkPixbuf.Colorspace.RGB,
                self._browse_pixbuf.get_has_alpha(),
                8,
                dest_width,
                dest_height,
            )

            if self._pixbuf is None:
                return

            self._pixbuf.fill(0x0)

            self._browse_pixbuf.scale(
                self._pixbuf,
                0,
                0,
                dest_width,
                dest_height,
                0,
                0,
                scale,
                scale,
                GdkPixbuf.InterpType.BILINEAR,
            )

            self._generate_color_shifted_pixbuf()

            if self._scale == 1.0:
                scale_to_80 = float(
                    min(
                        (self._pixbuf.get_width() * 0.8 / self._base_width),
                        (self._pixbuf.get_height() * 0.8 / self._base_height),
                    )
                )
                scale_to_image = float(
                    min(
                        (dest_width / self._base_width),
                        (dest_height / self._base_height),
                    )
                )
                crop_scale = float(min(scale_to_80, scale_to_image))

                self._crop.width = int(crop_scale * self._base_width / scale)
                self._crop.height = int(crop_scale * self._base_height / scale)
                self._crop.x = int(
                    (self._browse_pixbuf.get_width() - self._crop.width) / 2
                )
                self._crop.y = int(
                    (self._browse_pixbuf.get_height() - self._crop.height) / 2
                )

            self._scale = scale
            self._image.x = int((allocated_width - dest_width) / 2)
            self._image.y = int((allocated_height - dest_height) / 2)
            self._image.width = dest_width
            self._image.height = dest_height

    def _crop_to_widget(self) -> Gdk.Rectangle:
        crop = Gdk.Rectangle()
        crop.x = int(self._image.x + self._crop.x * self._scale)
        crop.y = int(self._image.y + self._crop.y * self._scale)
        crop.width = int(self._crop.width * self._scale)
        crop.height = int(self._crop.height * self._scale)
        return crop

    def _update_cursor(self, x_coord: float, y_coord: float) -> None:
        region = self._active_region
        if self._active_region == Loc.OUTSIDE:
            crop = self._crop_to_widget()
            region = self._find_location(crop, x_coord, y_coord)

        if region == Loc.TOP_LEFT:
            cursor_name = "nw-resize"
        elif region == Loc.TOP:
            cursor_name = "n-resize"
        elif region == Loc.TOP_RIGHT:
            cursor_name = "ne-resize"
        elif region == Loc.LEFT:
            cursor_name = "w-resize"
        elif region == Loc.INSIDE:
            cursor_name = "all-scroll"
        elif region == Loc.RIGHT:
            cursor_name = "e-resize"
        elif region == Loc.BOTTOM_LEFT:
            cursor_name = "sw-resize"
        elif region == Loc.BOTTOM:
            cursor_name = "s-resize"
        elif region == Loc.BOTTOM_RIGHT:
            cursor_name = "se-resize"
        else:  # Loc.OUTSIDE
            cursor_name = "default"

        if cursor_name is not self._current_cursor:
            self.set_cursor_from_name(cursor_name)
            self._current_cursor = cursor_name

    @staticmethod
    def _eval_radial_line(
        center_x: float, center_y: float, bounds_x: int, bounds_y: int, user_x: int
    ) -> int:
        slope_y = float(bounds_y - center_y)
        slope_x = bounds_x - center_x
        if slope_y == 0 or slope_x == 0:
            # Prevent division by zero
            return 0

        decision_slope = slope_y / slope_x
        decision_intercept = -float(decision_slope * bounds_x)
        return int(decision_slope * user_x + decision_intercept)

    def _find_location(
        self, rect: Gdk.Rectangle, x_coord: float, y_coord: float
    ) -> Loc:
        # pylint: disable=line-too-long
        location = [
            [Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE],
            [Loc.OUTSIDE, Loc.TOP_LEFT, Loc.TOP, Loc.TOP_RIGHT, Loc.OUTSIDE],
            [Loc.OUTSIDE, Loc.LEFT, Loc.INSIDE, Loc.RIGHT, Loc.OUTSIDE],
            [Loc.OUTSIDE, Loc.BOTTOM_LEFT, Loc.BOTTOM, Loc.BOTTOM_RIGHT, Loc.OUTSIDE],
            [Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE, Loc.OUTSIDE],
        ]
        # pylint: enable=line-too-long

        x_range = self._find_range(x_coord, rect.x, rect.x + rect.width)
        y_range = self._find_range(y_coord, rect.y, rect.y + rect.height)

        return location[y_range][x_range]

    @staticmethod
    def _find_range(coord: float, min_v: int, max_v: int) -> Range:
        tolerance = 12
        if coord < min_v - tolerance:
            return Range.BELOW
        if coord <= min_v + tolerance:
            return Range.LOWER
        if coord < max_v - tolerance:
            return Range.BETWEEN
        if coord <= max_v + tolerance:
            return Range.UPPER
        return Range.ABOVE

    def _generate_color_shifted_pixbuf(self) -> None:
        # pylint: disable=no-member
        assert self._pixbuf
        surface = cairo.ImageSurface(
            cairo.Format.ARGB32, self._pixbuf.get_width(), self._pixbuf.get_height()
        )
        context = cairo.Context(surface)
        # pylint: enable=no-member

        Gdk.cairo_set_source_pixbuf(context, self._pixbuf, 0, 0)
        context.paint()

        context.rectangle(0, 0, 1, 1)
        context.set_source_rgba(0, 0, 0, 0.5)
        context.paint()

        surface = context.get_target()
        assert isinstance(surface, cairo.ImageSurface)
        self._color_shifted_pixbuf = Gdk.pixbuf_get_from_surface(
            surface, 0, 0, surface.get_width(), surface.get_height()
        )



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.gtk.account_page import AccountPage
from gajim.gtk.chat_list import ChatList
from gajim.gtk.chat_page import ChatPage

PageT = ChatPage | AccountPage


class MainStack(Gtk.Stack):
    __gtype_name__ = "MainStack"

    def __init__(self) -> None:
        Gtk.Stack.__init__(self)

        self.add_named(Gtk.Box(), "empty")

        self._chat_page = ChatPage()
        self._chat_page.connect("chat-selected", self._on_chat_selected)
        self.add_named(self._chat_page, "chats")
        self.add_named(AccountPage(), "account")

    def _get_account_page(self) -> AccountPage:
        return cast(AccountPage, self.get_child_by_name("account"))

    def remove_account_page(self, account: str) -> None:
        account_page = self._get_account_page()
        if account_page.get_account() == account:
            account_page.set_account(None)
            self.set_visible_child_name("empty")

    def remove_chats_for_account(self, account: str) -> None:
        self._chat_page.remove_chats_for_account(account)

    def get_visible_page_name(self) -> str | None:
        return self.get_visible_child_name()

    def show_activity_page(self, context_id: str | None = None) -> None:
        self.set_visible_child_name("chats")
        self._chat_page.show_activity_page(context_id)

    def show_chats(self, workspace_id: str) -> None:
        self._chat_page.show_workspace_chats(workspace_id)
        self.set_visible_child_name("chats")

    def show_chat_page(self) -> None:
        self.set_visible_child_name("chats")

    def show_account(self, account: str) -> None:
        account_page = self._get_account_page()
        account_page.set_account(account)
        self.set_visible_child_name("account")

    def get_chat_page(self) -> ChatPage:
        chat_page = self.get_child_by_name("chats")
        assert isinstance(chat_page, ChatPage)
        return chat_page

    def _on_chat_selected(
        self, _chat_list: ChatList, _workspace_id: str, _account: str, _jid: JID
    ) -> None:
        self.set_visible_child_name("chats")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0100: Gateway Interaction

from __future__ import annotations

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import types
from gajim.common.events import AgentRemoved
from gajim.common.events import GatewayPromptReceived
from gajim.common.modules.base import BaseModule


class Gateway(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

    def unsubscribe(self, agent: str) -> None:
        if not app.account_is_available(self._account):
            return
        iq = nbxmpp.Iq('set', Namespace.REGISTER, to=agent)
        iq.setQuery().setTag('remove')

        self._con.connection.SendAndCallForResponse(
            iq, self._on_unsubscribe_result)
        self._con.get_module('Roster').delete_item(agent)

    def _on_unsubscribe_result(self,
                               _nbxmpp_client: types.NBXMPPClient,
                               stanza: nbxmpp.Protocol
                               ) -> None:
        if not nbxmpp.isResultNode(stanza):
            self._log.info('Error: %s', stanza.getError())
            return

        agent = stanza.getFrom().bare
        jid_list: list[JID] = []
        for contact in self._client.get_module('Roster').iter_contacts():
            if contact.jid.domain == agent:
                jid_list.append(contact.jid)
                self._log.info(
                    'Removing contact %s due to unregistered transport %s',
                    contact.jid,
                    agent)
                self._con.get_module('Presence').unsubscribe(contact.jid)

                # Transport contacts can't have 2 resources
                if contact.jid in app.to_be_removed[self._account]:
                    # This way we'll really remove it
                    app.to_be_removed[self._account].remove(contact.jid)

        app.ged.raise_event(
            AgentRemoved(conn=self._con,
                         agent=agent,
                         jid_list=jid_list))

    def request_gateway_prompt(self,
                               jid: str,
                               prompt: str | None = None
                               ) -> None:
        typ_ = 'get'
        if prompt:
            typ_ = 'set'
        iq = nbxmpp.Iq(typ=typ_, to=jid)
        query = iq.addChild(name='query', namespace=Namespace.GATEWAY)
        if prompt:
            query.setTagData('prompt', prompt)
        self._con.connection.SendAndCallForResponse(iq, self._on_prompt_result)

    def _on_prompt_result(self,
                          _nbxmpp_client: types.NBXMPPClient,
                          stanza: Iq
                          ) -> None:
        jid = str(stanza.getFrom())
        fjid = stanza.getFrom().bare
        resource = stanza.getFrom().resource

        query = stanza.getTag('query')
        if query is not None:
            desc = query.getTagData('desc')
            prompt = query.getTagData('prompt')
            prompt_jid = query.getTagData('jid')
        else:
            desc = None
            prompt = None
            prompt_jid = None

        app.ged.raise_event(
            GatewayPromptReceived(conn=self._con,
                                  fjid=fjid,
                                  jid=jid,
                                  resource=resource,
                                  desc=desc,
                                  prompt=prompt,
                                  prompt_jid=prompt_jid,
                                  stanza=stanza))



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _

from gajim.gtk.avatar import make_workspace_avatar
from gajim.gtk.avatar_selector import AvatarSelector
from gajim.gtk.builder import get_builder
from gajim.gtk.const import DEFAULT_WORKSPACE_COLOR
from gajim.gtk.util.styling import make_rgba
from gajim.gtk.util.styling import rgba_to_float
from gajim.gtk.window import GajimAppWindow


class WorkspaceDialog(GajimAppWindow):
    def __init__(self, workspace_id: str | None = None) -> None:
        GajimAppWindow.__init__(
            self,
            name="WorkspaceDialog",
            title=_("Workspace Settings"),
            default_width=500,
            default_height=600,
            add_window_padding=True,
            header_bar=True,
        )

        self._workspace_id = workspace_id

        self._ui = get_builder("workspace_dialog.ui")
        self.set_child(self._ui.box)

        self._avatar_selector = AvatarSelector()
        self._avatar_selector.set_size_request(400, 300)
        self._ui.image_box.append(self._avatar_selector)

        name: str = _("My Workspace")
        color: str | None = None
        self._avatar_sha: str | None = None

        if app.settings.get_workspace_count() == 1:
            # Don't allow to remove last workspace
            self._ui.remove_workspace_button.set_sensitive(False)

        if workspace_id is None:
            # This is a new workspace
            self._ui.remove_workspace_button.set_sensitive(False)
        else:
            name = app.settings.get_workspace_setting(workspace_id, "name")
            color = app.settings.get_workspace_setting(workspace_id, "color")
            self._avatar_sha = app.settings.get_workspace_setting(
                workspace_id, "avatar_sha"
            )
            if self._avatar_sha == "":
                self._avatar_sha = None

        rgba = make_rgba(color or DEFAULT_WORKSPACE_COLOR)
        if self._avatar_sha is not None:
            self._ui.image_switch.set_active(True)
            self._ui.style_stack.set_visible_child_name("image")

        self._ui.entry.set_text(name)

        self._ui.color_dialog_button.set_dialog(Gtk.ColorDialog())
        self._ui.color_dialog_button.set_rgba(rgba)

        self._update_avatar()

        self._connect(
            self._ui.remove_workspace_button, "clicked", self._on_remove_workspace
        )
        self._connect(
            self._ui.image_switch, "notify::active", self._on_image_switch_toggled
        )
        self._connect(self._ui.color_dialog_button, "notify::rgba", self._on_color_set)
        self._connect(self._ui.cancel_button, "clicked", self._on_cancel)
        self._connect(self._ui.save_button, "clicked", self._on_save)
        self._connect(self._ui.entry, "notify::text", self._on_text_changed)

        self.set_default_widget(self._ui.save_button)

    def _on_remove_workspace(self, _button: Gtk.Button) -> None:
        assert self._workspace_id is not None
        app.window.remove_workspace(self._workspace_id)
        self.close()

    def _on_cancel(self, _button: Gtk.Button) -> None:
        self.close()

    def _on_color_set(self, _button: Gtk.ColorDialogButton, *args: Any) -> None:
        self._update_avatar()

    def _on_text_changed(self, entry: Gtk.Entry, _param: Any) -> None:
        self._ui.save_button.set_sensitive(bool(entry.get_text()))
        self._update_avatar()

    def _on_image_switch_toggled(self, switch: Gtk.Switch, *args: Any) -> None:
        self._avatar_selector.reset()
        if switch.get_active():
            self._ui.style_stack.set_visible_child_name("image")
            if self._workspace_id is not None and app.settings.workspace_exists(
                self._workspace_id
            ):
                self._avatar_sha = app.settings.get_workspace_setting(
                    self._workspace_id, "avatar_sha"
                )
                if self._avatar_sha == "":
                    self._avatar_sha = None
        else:
            self._ui.style_stack.set_visible_child_name("color")
            self._avatar_sha = None

        self._update_avatar()

    def _update_avatar(self) -> None:
        name = self._ui.entry.get_text()
        rgba = self._ui.color_dialog_button.get_rgba()
        scale = self.get_scale_factor()
        if self._avatar_sha is not None:
            assert self._workspace_id is not None
            texture = app.app.avatar_storage.get_workspace_texture(
                self._workspace_id, AvatarSize.WORKSPACE_EDIT, scale
            )
        else:
            texture = make_workspace_avatar(
                name, rgba_to_float(rgba), AvatarSize.WORKSPACE_EDIT, scale
            )

        self._ui.preview.set_pixel_size(AvatarSize.WORKSPACE_EDIT)
        self._ui.preview.set_from_paintable(texture)

    def _get_avatar_data(self) -> bytes | None:
        if not self._avatar_selector.get_prepared():
            return None

        success, data, _wid, _hei = self._avatar_selector.get_avatar_bytes()
        if not success:
            return None

        return data

    def _on_save(self, _button: Gtk.Button) -> None:
        name = self._ui.entry.get_text()
        rgba = self._ui.color_dialog_button.get_rgba()
        use_image = self._ui.image_switch.get_active()
        if use_image:
            data = self._get_avatar_data()
            if data is not None:
                self._avatar_sha = app.app.avatar_storage.save_avatar(data)

        if self._workspace_id is not None and app.settings.workspace_exists(
            self._workspace_id
        ):
            app.settings.set_workspace_setting(self._workspace_id, "name", name)
            app.settings.set_workspace_setting(
                self._workspace_id, "color", rgba.to_string()
            )
            if self._avatar_sha is None:
                app.settings.set_workspace_setting(self._workspace_id, "avatar_sha", "")
            else:
                app.settings.set_workspace_setting(
                    self._workspace_id, "avatar_sha", self._avatar_sha
                )

            app.window.update_workspace(self._workspace_id)
            self.close()
            return

        workspace_id = app.settings.add_workspace(name)
        app.settings.set_workspace_setting(workspace_id, "color", rgba.to_string())
        if self._avatar_sha is not None:
            app.settings.set_workspace_setting(
                workspace_id, "avatar_sha", self._avatar_sha
            )

        app.window.add_workspace(workspace_id)
        self.close()

    def _cleanup(self) -> None:
        # https://gitlab.gnome.org/GNOME/gtk/-/issues/7992
        self._ui.entry.props.enable_emoji_completion = False



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import datetime as dt
import json
import logging

import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

from gajim.common import events
from gajim.common.storage.base import AlchemyStorage
from gajim.common.storage.base import json_decoder
from gajim.common.storage.base import with_session
from gajim.common.storage.events import models as mod
from gajim.common.types import ChatContactT

EventStorageEventT = (
    events.MUCNicknameChanged
    | events.MUCRoomConfigChanged
    | events.MUCRoomConfigFinished
    | events.MUCRoomPresenceError
    | events.MUCRoomKicked
    | events.MUCRoomDestroyed
    | events.MUCRoomVoiceRequestError
    | events.MUCUserJoined
    | events.MUCUserLeft
    | events.MUCUserRoleChanged
    | events.MUCUserAffiliationChanged
    | events.MUCUserStatusShowChanged
    | events.MUCUserHatsChanged
    | events.MUCAffiliationChanged
)

EVENT_CLASSES: dict[str, Any] = {
    "muc-nickname-changed": events.MUCNicknameChanged,
    "muc-room-config-changed": events.MUCRoomConfigChanged,
    "muc-room-config-finished": events.MUCRoomConfigFinished,
    "muc-room-presence-error": events.MUCRoomPresenceError,
    "muc-room-kicked": events.MUCRoomKicked,
    "muc-room-destroyed": events.MUCRoomDestroyed,
    "muc-room-voice-request-error": events.MUCRoomVoiceRequestError,
    "muc-user-joined": events.MUCUserJoined,
    "muc-user-left": events.MUCUserLeft,
    "muc-user-role-changed": events.MUCUserRoleChanged,
    "muc-user-affiliation-changed": events.MUCUserAffiliationChanged,
    "muc-user-status-show-changed": events.MUCUserStatusShowChanged,
    "muc-user-hats-changed": events.MUCUserHatsChanged,
    "room-affiliation-changed": events.MUCAffiliationChanged,
}

log = logging.getLogger("gajim.c.storage.events")


class EventStorage(AlchemyStorage):
    def __init__(self):
        AlchemyStorage.__init__(
            self,
            log,
            None,
        )

    def _create_table(self, session: Session, engine: Engine) -> None:
        mod.Base.metadata.create_all(engine)
        session.execute(sa.text("PRAGMA user_version=1"))

    def _migrate(self) -> None:
        pass

    @with_session
    def store(
        self, session: Session, contact: ChatContactT, event_: EventStorageEventT
    ) -> None:
        event = mod.Event(
            account=contact.account,
            jid=contact.jid,
            event=event_.name,
            timestamp=event_.timestamp,
            data=event_.serialize(),
        )

        session.add(event)

    @with_session
    def load(
        self,
        session: Session,
        contact: ChatContactT,
        before: bool,
        timestamp_: float,
        n_lines: int,
    ) -> list[events.ApplicationEvent]:
        timestamp = dt.datetime.fromtimestamp(timestamp_, dt.UTC)

        stmt = sa.select(mod.Event).where(
            mod.Event.account == contact.account, mod.Event.jid == contact.jid
        )

        if before:
            stmt = stmt.where(mod.Event.timestamp < timestamp).order_by(
                sa.desc(mod.Event.timestamp)
            )
        else:
            stmt = stmt.where(mod.Event.timestamp > timestamp).order_by(
                mod.Event.timestamp
            )

        stmt = stmt.limit(n_lines)

        event_list: list[events.ApplicationEvent] = []

        for event_row in session.scalars(stmt).all():
            event_class = EVENT_CLASSES[event_row.event]
            data = json.loads(event_row.data, object_hook=json_decoder)
            context_id = data.pop("context_id")
            event_ = event_class(**data, timestamp=event_row.timestamp)
            event_.context_id = context_id
            event_list.append(event_)

        return event_list
